// (c)2013 by Louis Coder. Implementation of a polynomial 2- and 3-SAT solver (at least I suppose that).

// ************************************
// *** ACTUAL POLYNOMIAL SOLVER >>> ***
// ************************************

typedef				unsigned __int64			u64;	// this var type u64 can hold a very large positive integer number

u64					Complexity1;						// for informing the user only; counts i loops in polynomial solver (in procedure Does_CNF_Have_A_Solution())
u64					Complexity2;						// for informing the user only; counts j loops in polynomial solver (in procedure Does_CNF_Have_A_Solution())
u64					Complexity3;						// for informing the user only; counts k loops in polynomial solver (in procedure Does_CNF_Have_A_Solution())
u64					Complexity4;						// for informing the user only; counts total loop iteration count in IsInConflict(), which is possibly called in the i, j and k loops
u64					ComplexityMax = 0;					// for informing the user only

int					DigitNumberOld = 0;					// used to know when to (re-)initialize PossibleClauses[][]
int					DigitNumber;						// the clause size of any clause in the CNF, examples: 5 for e.g. "0-00-" or 7 for e.g. "1-0-0--"
int					ClauseNumber;						// number of clauses in the CNF
char*				CNF;								// pointer to a char string containing the CNF

int					SAT_TYPE_OLD = 0;					// used to detect changes
extern int			SAT_TYPE;							// 2 for 2-SAT or 3 for 3-SAT ('extern' as originally defined in Polynomial_3SAT_Solver.cpp)

int					CombinationCount;					// count of possible 0/1 combinations (see also function 'Does_CNF_Have_A_Solution()')

#define				POSSIBLE_CLAUSE_NUMBER_MAX	(18500)	// highest PossibleClauseNumber supported by the solver (value gained through tests)

int					PossibleClauseNumber;				// depends from DigitNumber and the choice 2-SAT or 3-SAT
char				PossibleClauses[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX];	// see PossibleClauses_Create()

// The following data might be used only to produce debug output.
// That debug output shall help the user to understand the algorithm.

#define				CT_DIGIT_NUMBER_MAX			(12)	// 2^12 = 4096 = CT_LINE_NUMBER_MAX; do not use/display the ClauseTable if DigitNumber is greater than this!
#define				CT_LINE_NUMBER_MAX			(4096)	// how many lines the ClauseTable can maximal have
#define				CT_COLUMN_NUMBER_MAX		(1024)	// how many columns the ClauseTable can maximal have

int					CTLineNumber = 0;					// the current count of lines in the ClauseTable
int					CTColumnNumber;						// the current count of columns in the ClauseTable
char				CT[CT_LINE_NUMBER_MAX][CT_COLUMN_NUMBER_MAX][CT_DIGIT_NUMBER_MAX + 1];	// contains Clauses; + 1 for null terminator

// **************************************************************

bool IsInConflict(char* Clause1, char* Clause2)
{
	// For instance, IS in conflict:
	// "0-0-" and "1-0-".
	// NOT in conflict:
	// "0-0-" and "-00-" or
	// "0-0-" and "-0-0" or
	// "0-0-" and "-10-" is also NOT in conflict.

	for (int m = 0; m < DigitNumber; m ++)
	{
		if ((Clause1[m] == '0' && Clause2[m] == '1') ||
			(Clause1[m] == '1' && Clause2[m] == '0'))
		{
			Complexity4 += m + 1;
			return true;
		}
	}

	Complexity4 += DigitNumber;
	return false;
}

bool IsEqual(char* Clause1, char* Clause2)
{
	// checks if the two char strings are identical.
	// DigitNumber is a global var initialized in
	// CNF_Solve_Polynomial().

	for (int m = 0; m < DigitNumber; m ++)
		if ((Clause1[m] != Clause2[m]))
			return false;
	
	return true;
}

bool ExistsInCNF(char* Clause)
{
	// CNF is a global pointer initialized in
	// CNF_Solve_Polynomial(). This pointer
	// points to a char string that contains
	// the CNF in Clause-notation.
	// Each clause is DigitNumber digits wide.
	// So x * DigitNumber points to the beginning
	// of the x-th clause (x is 0-based).
	
	for (int i = 0; i < ClauseNumber; i ++)
		if (IsEqual(Clause, &CNF[i * DigitNumber]))
			return true;

	return false;
}

bool ExistsInCNFCache[POSSIBLE_CLAUSE_NUMBER_MAX];

void ExistsInCNF_Cache() // must be called for every CNF
{
	for (int i = 0; i < PossibleClauseNumber; i ++)
		ExistsInCNFCache[i] = ExistsInCNF(PossibleClauses[i]);
}

bool ExistsInCNF(int PossibleClauseIndex)
{
	return ExistsInCNFCache[PossibleClauseIndex];
}

// **************************************************************

void ShowClause(char* Clause, bool NewLine = true)
{
	// for debugging.
	// You might use it like this:
	// ShowClause(PossibleClauses[x]);
	// or
	// ShowClause(&CNF[ClauseIndex * DigitNumber]);.

	if (Clause == NULL)
		return;

	for (int m = 0; m < DigitNumber; m ++)
		printf("%c", Clause[m]);
	if (NewLine)
		printf("\n");
}

// **************************************************************

void PossibleClauses_Create()
{
	PossibleClauseNumber = 0;

	if (SAT_TYPE == 2)
	{
		// if DigitNumber = 3:
		// 00-
		// 01-
		// 10-
		// 11-
		// 0-0
		// 0-1
		// 1-0
		// 1-1
		// -00
		// -01
		// -10
		// -11
		// PossibleClauseNumber = (3 * 4) = 12

		for (int D1 = 0; D1 < DigitNumber - 1; D1 ++) // position of first 0 or 1 digit in PossibleClause
		{
			for (int D2 = D1 + 1; D2 < DigitNumber; D2 ++) // position of second 0 or 1 digit in PossibleClause
			{
				for (int C = 0; C < 4; C ++) // CombinationIndex: 00, 01, 10, 11
				{
					// we arrive here O(DigitNumber^2) times!
					// So there are O(DigitNumber^2) PossibleClauses!

					for (int m = 0; m < DigitNumber; m ++)
						PossibleClauses[PossibleClauseNumber][m] = '-';

					PossibleClauses[PossibleClauseNumber][D1] = (C & 2 ? '1' : '0');
					PossibleClauses[PossibleClauseNumber][D2] = (C & 1 ? '1' : '0');
					PossibleClauses[PossibleClauseNumber][DigitNumber] = '\0';

					PossibleClauseNumber ++;

					if (PossibleClauseNumber > POSSIBLE_CLAUSE_NUMBER_MAX) // if this happens, try to increase POSSIBLE_CLAUSE_NUMBER_MAX and re-compile
					{
						printf("Warning: PossibleClauseNumber too large in PossibleClauses_Create() !");
						getchar();
					}
				}
			}
		}
	}

	if (SAT_TYPE == 3)
	{
		// if DigitNumber = 4:
		// 000-
		// 001-
		// 010-
		// 011-
		// 100-
		// 101-
		// 110-
		// 111-
		// 00-0
		// 00-1
		// 01-0
		// 01-1
		// 10-0
		// 10-1
		// 11-0
		// 11-1
		// 0-00
		// 0-01
		// 0-10
		// 0-11
		// 1-00
		// 1-01
		// 1-10
		// 1-11
		// -000
		// -001
		// -010
		// -011
		// -100
		// -101
		// -110
		// -111
		// PossibleClauseNumber = (4 * 8) = 24

		for (int D1 = 0; D1 < DigitNumber - 2; D1 ++) // position of first 0 or 1 digit in PossibleClause
		{
			for (int D2 = D1 + 1; D2 < DigitNumber - 1; D2 ++) // position of second 0 or 1 digit in PossibleClause
			{
				for (int D3 = D2 + 1; D3 < DigitNumber; D3 ++) // position of third 0 or 1 digit in PossibleClause
				{
					for (int C = 0; C < 8; C ++) // 000, 001, 010, 011, 100, 101, 110, 111
					{
						// we arrive here O(DigitNumber^3) times!
						// So there are O(DigitNumber^3) PossibleClauses!

						for (int m = 0; m < DigitNumber; m ++)
							PossibleClauses[PossibleClauseNumber][m] = '-';

						PossibleClauses[PossibleClauseNumber][D1] = (C & 4 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][D2] = (C & 2 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][D3] = (C & 1 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][DigitNumber] = '\0';

						PossibleClauseNumber ++;

						if (PossibleClauseNumber > POSSIBLE_CLAUSE_NUMBER_MAX) // if this happens, try to increase POSSIBLE_CLAUSE_NUMBER_MAX and re-compile
						{
							printf("Warning: PossibleClauseNumber too large in PossibleClauses_Create() !");
							getchar();
						}
					}
				}
			}
		}
	}
}



// *** NOT REQUIRED FOR ACTUAL SOLVER BUT FOR DEBUGGING/VISUALIZATION ONLY >>> ***

int GetPossibleClauseIndex(char* PossibleClause) // rather slow - use for debugging only!
{
	for (int i = 0; i < PossibleClauseNumber; i ++)
		if (PossibleClause[0] == PossibleClauses[i][0] &&
			PossibleClause[1] == PossibleClauses[i][1]) // quick incomplete check to increase speed
			if (IsEqual(PossibleClause, PossibleClauses[i]))
				return i;

	printf("ERROR in GetPossibleClauseIndex(): argument is not a PossibleClause or ClauseTable has not been created correctly !\n");
	getchar();
	return -1;
}

void ClauseTable_CreateDynamically() // fills CT[][] and sets CTLineNumber and CTColumnNumber
{
	CTLineNumber = 0;

	if (DigitNumber > CT_DIGIT_NUMBER_MAX)
	{
		printf("ERROR: ClauseTable_CreateDynamically() can only\n");
		printf("       be used for a DigitNumber <= %d !\n", CT_DIGIT_NUMBER_MAX);
		printf("       You tried to use it for DigitNumber = %d.\n", DigitNumber);
		getchar();
	}

	char CurrentLine[DIGIT_NUMBER_MAX];

	for (int i = 0; i < DigitNumber; i ++)
		CurrentLine[i] = '0';

	// CurrentLine is initially (for DigitNumber = 5):
	// 00000.
	// In next while-iteration, it is:
	// 00001.
	// Then
	// 00010,
	// 00011,
	// 00100,
	// 00101,
	// 00110,
	// 00111, ...
	// Notice that this is EXPONENTIAL, but used for debugging/visualization ONLY!

	while (true)
	{
		for (int i = 0; i < CT_COLUMN_NUMBER_MAX; i ++)
			for (int m = 0; m < CT_DIGIT_NUMBER_MAX; m ++)
				CT[CTLineNumber][i][m] = '\0'; // clear

		int CTColumnIndex = 0;

		if (SAT_TYPE == 2)
		{
			for (int D1 = 0; D1 < DigitNumber - 1; D1 ++)
			{
				for (int D2 = D1 + 1; D2 < DigitNumber; D2 ++)
				{
					for (int m = 0; m < DigitNumber; m ++)
						CT[CTLineNumber][CTColumnIndex][m] = '-';
					
					CT[CTLineNumber][CTColumnIndex][D1] = CurrentLine[D1];
					CT[CTLineNumber][CTColumnIndex][D2] = CurrentLine[D2];
					CT[CTLineNumber][CTColumnIndex][DigitNumber] = '\0';

					CTColumnIndex ++;
				}
			}
		}

		if (SAT_TYPE == 3)
		{
			for (int D1 = 0; D1 < DigitNumber - 2; D1 ++)
			{
				for (int D2 = D1 + 1; D2 < DigitNumber - 1; D2 ++)
				{
					for (int D3 = D2 + 1; D3 < DigitNumber; D3 ++)
					{
						for (int m = 0; m < DigitNumber; m ++)
							CT[CTLineNumber][CTColumnIndex][m] = '-';

						CT[CTLineNumber][CTColumnIndex][D1] = CurrentLine[D1];
						CT[CTLineNumber][CTColumnIndex][D2] = CurrentLine[D2];
						CT[CTLineNumber][CTColumnIndex][D3] = CurrentLine[D3];
						CT[CTLineNumber][CTColumnIndex][DigitNumber] = '\0';

						CTColumnIndex ++;
					}
				}
			}
		}

		CTColumnNumber = CTColumnIndex;
		CTLineNumber ++;

		if (CTColumnNumber > CT_COLUMN_NUMBER_MAX)
		{
			printf("CTColumnNumber overflow !");
			getchar();
		}
		if (CTLineNumber > CT_LINE_NUMBER_MAX)
		{
			printf("CTLineNumber overflow !");
			getchar();
		}

		for (int m = DigitNumber - 1; m >= 0; m --)
			if (CurrentLine[m] == '0')
			{
				CurrentLine[m] = '1';
				goto NextLine;
			}
			else if (CurrentLine[m] == '1')
				CurrentLine[m] = '0';

		break;
NextLine:;
	}
}

void ClauseTable_Show()
{ // shows the whole ClauseTable (in the Windows console or, if specially compiled therefore, in the Linux shell)

	printf("CLAUSE TABLE >>>\n\n");

	for (int i = 0; i < CTLineNumber; i ++) // loop through the ClauseTable lines
	{
		// show header that tells which PossibleClauseIndices the clauses in that ClauseTable column have
		if (i == 0)
		{
			printf("['C (X)' means clause C has PossibleClauseIndex X.]\n\n");

			for (int j = 0; j < CTColumnNumber; j ++) // loop through the ClauseTable columns
			{
				static char S[1024 * 1024];
				S[0] = '\0'; // important for sprintf()

				sprintf(S, "%d to %d...", (j + 0) * CombinationCount, (j + 1) * CombinationCount - 1);

				printf("%s", S);

				for (int m = DigitNumber + 6 + 20 - (int)strlen(S); m >= 0; m --)
					printf(" ");
			}
			printf("\n\n");
		}

		// show the clauses and their PossibleClauseIndex
		for (int j = 0; j < CTColumnNumber; j ++)
		{
			ShowClause(CT[i][j], false);

			int PossibleClauseIndex = GetPossibleClauseIndex(CT[i][j]);

			if (PossibleClauseIndex < 10)
				printf(" (%d)  ", PossibleClauseIndex);
			else
			if (PossibleClauseIndex < 100)
				printf(" (%d) ", PossibleClauseIndex);
			else
				printf(" (%d)", PossibleClauseIndex);

			static char S[1024 * 1024];
			S[0] = '\0'; // important for sprintf()/strlen()

			printf("%s", S);

			for (int m = 20 - (int)strlen(S); m >= 0; m --)
				printf(" ");
		}

		printf("\n");
	}

	printf("\n<<< END OF CLAUSE TABLE\n\n");
}

bool M(char* Clause) // [clause] Missing in CNF to solve?
{
	return !ExistsInCNF(Clause);
}

int ClauseTableTest()
{
	// NOTE: this procedure shall show that you can really
	// determine if a CNF is solvable or not by using the
	// 'ClauseTable'.
	// The following implementation of the 'ClauseTable'
	// works for 2-SAT CNFs with a DigitNumber of 4, 5 or 6.
	// 
	// You could construct a 'ClauseTable' for 3-SAT as well,
	// there's no significant difference, except that the
	// size of the table (number of M("...") calls) changes.

	switch (DigitNumber)
	{
	case 4: // 2^4 = 16 ClauseTableLines, 6 ClauseTableColumns
		return
			(M("00--") && M("-00-") && M("--00") && M("0-0-") && M("-0-0") && M("0--0")) |
			(M("00--") && M("-00-") && M("--01") && M("0-0-") && M("-0-1") && M("0--1")) |
			(M("00--") && M("-01-") && M("--10") && M("0-1-") && M("-0-0") && M("0--0")) |
			(M("00--") && M("-01-") && M("--11") && M("0-1-") && M("-0-1") && M("0--1")) |
			(M("01--") && M("-10-") && M("--00") && M("0-0-") && M("-1-0") && M("0--0")) |
			(M("01--") && M("-10-") && M("--01") && M("0-0-") && M("-1-1") && M("0--1")) |
			(M("01--") && M("-11-") && M("--10") && M("0-1-") && M("-1-0") && M("0--0")) |
			(M("01--") && M("-11-") && M("--11") && M("0-1-") && M("-1-1") && M("0--1")) |
			(M("10--") && M("-00-") && M("--00") && M("1-0-") && M("-0-0") && M("1--0")) |
			(M("10--") && M("-00-") && M("--01") && M("1-0-") && M("-0-1") && M("1--1")) |
			(M("10--") && M("-01-") && M("--10") && M("1-1-") && M("-0-0") && M("1--0")) |
			(M("10--") && M("-01-") && M("--11") && M("1-1-") && M("-0-1") && M("1--1")) |
			(M("11--") && M("-10-") && M("--00") && M("1-0-") && M("-1-0") && M("1--0")) |
			(M("11--") && M("-10-") && M("--01") && M("1-0-") && M("-1-1") && M("1--1")) |
			(M("11--") && M("-11-") && M("--10") && M("1-1-") && M("-1-0") && M("1--0")) |
			(M("11--") && M("-11-") && M("--11") && M("1-1-") && M("-1-1") && M("1--1"));

	case 5: // 2^5 = 32 ClauseTableLines, 10 ClauseTableColumns
		return
			// 00000:
			(M("00---") && M("-00--") && M("--00-") && M("---00") && M("0-0--") && M("0--0-") && M("0---0") && M("-0-0-") && M("-0--0") && M("--0-0")) |
			// 00001:
			(M("00---") && M("-00--") && M("--00-") && M("---01") && M("0-0--") && M("0--0-") && M("0---1") && M("-0-0-") && M("-0--1") && M("--0-1")) |
			// 00010:
			(M("00---") && M("-00--") && M("--01-") && M("---10") && M("0-0--") && M("0--1-") && M("0---0") && M("-0-1-") && M("-0--0") && M("--0-0")) |
			// 00011:
			(M("00---") && M("-00--") && M("--01-") && M("---11") && M("0-0--") && M("0--1-") && M("0---1") && M("-0-1-") && M("-0--1") && M("--0-1")) |
			// 00100:
			(M("00---") && M("-01--") && M("--10-") && M("---00") && M("0-1--") && M("0--0-") && M("0---0") && M("-0-0-") && M("-0--0") && M("--1-0")) |
			// 00101:
			(M("00---") && M("-01--") && M("--10-") && M("---01") && M("0-1--") && M("0--0-") && M("0---1") && M("-0-0-") && M("-0--1") && M("--1-1")) |
			// 00110:
			(M("00---") && M("-01--") && M("--11-") && M("---10") && M("0-1--") && M("0--1-") && M("0---0") && M("-0-1-") && M("-0--0") && M("--1-0")) |
			// 00111:
			(M("00---") && M("-01--") && M("--11-") && M("---11") && M("0-1--") && M("0--1-") && M("0---1") && M("-0-1-") && M("-0--1") && M("--1-1")) |

			// 01000:
			(M("01---") && M("-10--") && M("--00-") && M("---00") && M("0-0--") && M("0--0-") && M("0---0") && M("-1-0-") && M("-1--0") && M("--0-0")) |
			// 01001:
			(M("01---") && M("-10--") && M("--00-") && M("---01") && M("0-0--") && M("0--0-") && M("0---1") && M("-1-0-") && M("-1--1") && M("--0-1")) |
			// 01010:
			(M("01---") && M("-10--") && M("--01-") && M("---10") && M("0-0--") && M("0--1-") && M("0---0") && M("-1-1-") && M("-1--0") && M("--0-0")) |
			// 01011:
			(M("01---") && M("-10--") && M("--01-") && M("---11") && M("0-0--") && M("0--1-") && M("0---1") && M("-1-1-") && M("-1--1") && M("--0-1")) |
			// 01100:
			(M("01---") && M("-11--") && M("--10-") && M("---00") && M("0-1--") && M("0--0-") && M("0---0") && M("-1-0-") && M("-1--0") && M("--1-0")) |
			// 01101:
			(M("01---") && M("-11--") && M("--10-") && M("---01") && M("0-1--") && M("0--0-") && M("0---1") && M("-1-0-") && M("-1--1") && M("--1-1")) |
			// 01110:
			(M("01---") && M("-11--") && M("--11-") && M("---10") && M("0-1--") && M("0--1-") && M("0---0") && M("-1-1-") && M("-1--0") && M("--1-0")) |
			// 01111:
			(M("01---") && M("-11--") && M("--11-") && M("---11") && M("0-1--") && M("0--1-") && M("0---1") && M("-1-1-") && M("-1--1") && M("--1-1")) |

			// 10000:
			(M("10---") && M("-00--") && M("--00-") && M("---00") && M("1-0--") && M("1--0-") && M("1---0") && M("-0-0-") && M("-0--0") && M("--0-0")) |
			// 10001:
			(M("10---") && M("-00--") && M("--00-") && M("---01") && M("1-0--") && M("1--0-") && M("1---1") && M("-0-0-") && M("-0--1") && M("--0-1")) |
			// 10010:
			(M("10---") && M("-00--") && M("--01-") && M("---10") && M("1-0--") && M("1--1-") && M("1---0") && M("-0-1-") && M("-0--0") && M("--0-0")) |
			// 10011:
			(M("10---") && M("-00--") && M("--01-") && M("---11") && M("1-0--") && M("1--1-") && M("1---1") && M("-0-1-") && M("-0--1") && M("--0-1")) |
			// 10100:
			(M("10---") && M("-01--") && M("--10-") && M("---00") && M("1-1--") && M("1--0-") && M("1---0") && M("-0-0-") && M("-0--0") && M("--1-0")) |
			// 10101:
			(M("10---") && M("-01--") && M("--10-") && M("---01") && M("1-1--") && M("1--0-") && M("1---1") && M("-0-0-") && M("-0--1") && M("--1-1")) |
			// 10110:
			(M("10---") && M("-01--") && M("--11-") && M("---10") && M("1-1--") && M("1--1-") && M("1---0") && M("-0-1-") && M("-0--0") && M("--1-0")) |
			// 10111:
			(M("10---") && M("-01--") && M("--11-") && M("---11") && M("1-1--") && M("1--1-") && M("1---1") && M("-0-1-") && M("-0--1") && M("--1-1")) |

			// 11000:
			(M("11---") && M("-10--") && M("--00-") && M("---00") && M("1-0--") && M("1--0-") && M("1---0") && M("-1-0-") && M("-1--0") && M("--0-0")) |
			// 11001:
			(M("11---") && M("-10--") && M("--00-") && M("---01") && M("1-0--") && M("1--0-") && M("1---1") && M("-1-0-") && M("-1--1") && M("--0-1")) |
			// 11010:
			(M("11---") && M("-10--") && M("--01-") && M("---10") && M("1-0--") && M("1--1-") && M("1---0") && M("-1-1-") && M("-1--0") && M("--0-0")) |
			// 11011:
			(M("11---") && M("-10--") && M("--01-") && M("---11") && M("1-0--") && M("1--1-") && M("1---1") && M("-1-1-") && M("-1--1") && M("--0-1")) |
			// 11100:
			(M("11---") && M("-11--") && M("--10-") && M("---00") && M("1-1--") && M("1--0-") && M("1---0") && M("-1-0-") && M("-1--0") && M("--1-0")) |
			// 11101:
			(M("11---") && M("-11--") && M("--10-") && M("---01") && M("1-1--") && M("1--0-") && M("1---1") && M("-1-0-") && M("-1--1") && M("--1-1")) |
			// 11110:
			(M("11---") && M("-11--") && M("--11-") && M("---10") && M("1-1--") && M("1--1-") && M("1---0") && M("-1-1-") && M("-1--0") && M("--1-0")) |
			// 11111:
			(M("11---") && M("-11--") && M("--11-") && M("---11") && M("1-1--") && M("1--1-") && M("1---1") && M("-1-1-") && M("-1--1") && M("--1-1"));

	case 6: // 2^6 = 64 ClauseTableLines, 15 ClauseTableColumns
		return
			// 000000:
			(M("00----") && M("-00---") && M("--00--") && M("---00-") && M("----00") && M("0-0---") && M("0--0--") && M("0---0-") && M("0----0") && M("-0-0--") && M("-0--0-") && M("-0---0") && M("--0-0-") && M("--0--0") && M("---0-0")) |
			// 000001:
			(M("00----") && M("-00---") && M("--00--") && M("---00-") && M("----01") && M("0-0---") && M("0--0--") && M("0---0-") && M("0----1") && M("-0-0--") && M("-0--0-") && M("-0---1") && M("--0-0-") && M("--0--1") && M("---0-1")) |
			// 000010:
			(M("00----") && M("-00---") && M("--00--") && M("---01-") && M("----10") && M("0-0---") && M("0--0--") && M("0---1-") && M("0----0") && M("-0-0--") && M("-0--1-") && M("-0---0") && M("--0-1-") && M("--0--0") && M("---0-0")) |
			// 000011:
			(M("00----") && M("-00---") && M("--00--") && M("---01-") && M("----11") && M("0-0---") && M("0--0--") && M("0---1-") && M("0----1") && M("-0-0--") && M("-0--1-") && M("-0---1") && M("--0-1-") && M("--0--1") && M("---0-1")) |
			// 000100:
			(M("00----") && M("-00---") && M("--01--") && M("---10-") && M("----00") && M("0-0---") && M("0--1--") && M("0---0-") && M("0----0") && M("-0-1--") && M("-0--0-") && M("-0---0") && M("--0-0-") && M("--0--0") && M("---1-0")) |
			// 000101:
			(M("00----") && M("-00---") && M("--01--") && M("---10-") && M("----01") && M("0-0---") && M("0--1--") && M("0---0-") && M("0----1") && M("-0-1--") && M("-0--0-") && M("-0---1") && M("--0-0-") && M("--0--1") && M("---1-1")) |
			// 000110:
			(M("00----") && M("-00---") && M("--01--") && M("---11-") && M("----10") && M("0-0---") && M("0--1--") && M("0---1-") && M("0----0") && M("-0-1--") && M("-0--1-") && M("-0---0") && M("--0-1-") && M("--0--0") && M("---1-0")) |
			// 000111:
			(M("00----") && M("-00---") && M("--01--") && M("---11-") && M("----11") && M("0-0---") && M("0--1--") && M("0---1-") && M("0----1") && M("-0-1--") && M("-0--1-") && M("-0---1") && M("--0-1-") && M("--0--1") && M("---1-1")) |

			// 001000:
			(M("00----") && M("-01---") && M("--10--") && M("---00-") && M("----00") && M("0-1---") && M("0--0--") && M("0---0-") && M("0----0") && M("-0-0--") && M("-0--0-") && M("-0---0") && M("--1-0-") && M("--1--0") && M("---0-0")) |
			// 001001:
			(M("00----") && M("-01---") && M("--10--") && M("---00-") && M("----01") && M("0-1---") && M("0--0--") && M("0---0-") && M("0----1") && M("-0-0--") && M("-0--0-") && M("-0---1") && M("--1-0-") && M("--1--1") && M("---0-1")) |
			// 001010:
			(M("00----") && M("-01---") && M("--10--") && M("---01-") && M("----10") && M("0-1---") && M("0--0--") && M("0---1-") && M("0----0") && M("-0-0--") && M("-0--1-") && M("-0---0") && M("--1-1-") && M("--1--0") && M("---0-0")) |
			// 001011:
			(M("00----") && M("-01---") && M("--10--") && M("---01-") && M("----11") && M("0-1---") && M("0--0--") && M("0---1-") && M("0----1") && M("-0-0--") && M("-0--1-") && M("-0---1") && M("--1-1-") && M("--1--1") && M("---0-1")) |
			// 001100:
			(M("00----") && M("-01---") && M("--11--") && M("---10-") && M("----00") && M("0-1---") && M("0--1--") && M("0---0-") && M("0----0") && M("-0-1--") && M("-0--0-") && M("-0---0") && M("--1-0-") && M("--1--0") && M("---1-0")) |
			// 001101:
			(M("00----") && M("-01---") && M("--11--") && M("---10-") && M("----01") && M("0-1---") && M("0--1--") && M("0---0-") && M("0----1") && M("-0-1--") && M("-0--0-") && M("-0---1") && M("--1-0-") && M("--1--1") && M("---1-1")) |
			// 001110:
			(M("00----") && M("-01---") && M("--11--") && M("---11-") && M("----10") && M("0-1---") && M("0--1--") && M("0---1-") && M("0----0") && M("-0-1--") && M("-0--1-") && M("-0---0") && M("--1-1-") && M("--1--0") && M("---1-0")) |
			// 001111:
			(M("00----") && M("-01---") && M("--11--") && M("---11-") && M("----11") && M("0-1---") && M("0--1--") && M("0---1-") && M("0----1") && M("-0-1--") && M("-0--1-") && M("-0---1") && M("--1-1-") && M("--1--1") && M("---1-1")) |

			// 010000:
			(M("01----") && M("-10---") && M("--00--") && M("---00-") && M("----00") && M("0-0---") && M("0--0--") && M("0---0-") && M("0----0") && M("-1-0--") && M("-1--0-") && M("-1---0") && M("--0-0-") && M("--0--0") && M("---0-0")) |
			// 010001:
			(M("01----") && M("-10---") && M("--00--") && M("---00-") && M("----01") && M("0-0---") && M("0--0--") && M("0---0-") && M("0----1") && M("-1-0--") && M("-1--0-") && M("-1---1") && M("--0-0-") && M("--0--1") && M("---0-1")) |
			// 010010:
			(M("01----") && M("-10---") && M("--00--") && M("---01-") && M("----10") && M("0-0---") && M("0--0--") && M("0---1-") && M("0----0") && M("-1-0--") && M("-1--1-") && M("-1---0") && M("--0-1-") && M("--0--0") && M("---0-0")) |
			// 010011:
			(M("01----") && M("-10---") && M("--00--") && M("---01-") && M("----11") && M("0-0---") && M("0--0--") && M("0---1-") && M("0----1") && M("-1-0--") && M("-1--1-") && M("-1---1") && M("--0-1-") && M("--0--1") && M("---0-1")) |
			// 010100:
			(M("01----") && M("-10---") && M("--01--") && M("---10-") && M("----00") && M("0-0---") && M("0--1--") && M("0---0-") && M("0----0") && M("-1-1--") && M("-1--0-") && M("-1---0") && M("--0-0-") && M("--0--0") && M("---1-0")) |
			// 010101:
			(M("01----") && M("-10---") && M("--01--") && M("---10-") && M("----01") && M("0-0---") && M("0--1--") && M("0---0-") && M("0----1") && M("-1-1--") && M("-1--0-") && M("-1---1") && M("--0-0-") && M("--0--1") && M("---1-1")) |
			// 010110:
			(M("01----") && M("-10---") && M("--01--") && M("---11-") && M("----10") && M("0-0---") && M("0--1--") && M("0---1-") && M("0----0") && M("-1-1--") && M("-1--1-") && M("-1---0") && M("--0-1-") && M("--0--0") && M("---1-0")) |
			// 010111:
			(M("01----") && M("-10---") && M("--01--") && M("---11-") && M("----11") && M("0-0---") && M("0--1--") && M("0---1-") && M("0----1") && M("-1-1--") && M("-1--1-") && M("-1---1") && M("--0-1-") && M("--0--1") && M("---1-1")) |

			// 011000:
			(M("01----") && M("-11---") && M("--10--") && M("---00-") && M("----00") && M("0-1---") && M("0--0--") && M("0---0-") && M("0----0") && M("-1-0--") && M("-1--0-") && M("-1---0") && M("--1-0-") && M("--1--0") && M("---0-0")) |
			// 011001:
			(M("01----") && M("-11---") && M("--10--") && M("---00-") && M("----01") && M("0-1---") && M("0--0--") && M("0---0-") && M("0----1") && M("-1-0--") && M("-1--0-") && M("-1---1") && M("--1-0-") && M("--1--1") && M("---0-1")) |
			// 011010:
			(M("01----") && M("-11---") && M("--10--") && M("---01-") && M("----10") && M("0-1---") && M("0--0--") && M("0---1-") && M("0----0") && M("-1-0--") && M("-1--1-") && M("-1---0") && M("--1-1-") && M("--1--0") && M("---0-0")) |
			// 011011:
			(M("01----") && M("-11---") && M("--10--") && M("---01-") && M("----11") && M("0-1---") && M("0--0--") && M("0---1-") && M("0----1") && M("-1-0--") && M("-1--1-") && M("-1---1") && M("--1-1-") && M("--1--1") && M("---0-1")) |
			// 011100:
			(M("01----") && M("-11---") && M("--11--") && M("---10-") && M("----00") && M("0-1---") && M("0--1--") && M("0---0-") && M("0----0") && M("-1-1--") && M("-1--0-") && M("-1---0") && M("--1-0-") && M("--1--0") && M("---1-0")) |
			// 011101:
			(M("01----") && M("-11---") && M("--11--") && M("---10-") && M("----01") && M("0-1---") && M("0--1--") && M("0---0-") && M("0----1") && M("-1-1--") && M("-1--0-") && M("-1---1") && M("--1-0-") && M("--1--1") && M("---1-1")) |
			// 011110:
			(M("01----") && M("-11---") && M("--11--") && M("---11-") && M("----10") && M("0-1---") && M("0--1--") && M("0---1-") && M("0----0") && M("-1-1--") && M("-1--1-") && M("-1---0") && M("--1-1-") && M("--1--0") && M("---1-0")) |
			// 011111:
			(M("01----") && M("-11---") && M("--11--") && M("---11-") && M("----11") && M("0-1---") && M("0--1--") && M("0---1-") && M("0----1") && M("-1-1--") && M("-1--1-") && M("-1---1") && M("--1-1-") && M("--1--1") && M("---1-1")) |


			// 100000:
			(M("10----") && M("-00---") && M("--00--") && M("---00-") && M("----00") && M("1-0---") && M("1--0--") && M("1---0-") && M("1----0") && M("-0-0--") && M("-0--0-") && M("-0---0") && M("--0-0-") && M("--0--0") && M("---0-0")) |
			// 100001:
			(M("10----") && M("-00---") && M("--00--") && M("---00-") && M("----01") && M("1-0---") && M("1--0--") && M("1---0-") && M("1----1") && M("-0-0--") && M("-0--0-") && M("-0---1") && M("--0-0-") && M("--0--1") && M("---0-1")) |
			// 100010:
			(M("10----") && M("-00---") && M("--00--") && M("---01-") && M("----10") && M("1-0---") && M("1--0--") && M("1---1-") && M("1----0") && M("-0-0--") && M("-0--1-") && M("-0---0") && M("--0-1-") && M("--0--0") && M("---0-0")) |
			// 100011:
			(M("10----") && M("-00---") && M("--00--") && M("---01-") && M("----11") && M("1-0---") && M("1--0--") && M("1---1-") && M("1----1") && M("-0-0--") && M("-0--1-") && M("-0---1") && M("--0-1-") && M("--0--1") && M("---0-1")) |
			// 100100:
			(M("10----") && M("-00---") && M("--01--") && M("---10-") && M("----00") && M("1-0---") && M("1--1--") && M("1---0-") && M("1----0") && M("-0-1--") && M("-0--0-") && M("-0---0") && M("--0-0-") && M("--0--0") && M("---1-0")) |
			// 100101:
			(M("10----") && M("-00---") && M("--01--") && M("---10-") && M("----01") && M("1-0---") && M("1--1--") && M("1---0-") && M("1----1") && M("-0-1--") && M("-0--0-") && M("-0---1") && M("--0-0-") && M("--0--1") && M("---1-1")) |
			// 100110:
			(M("10----") && M("-00---") && M("--01--") && M("---11-") && M("----10") && M("1-0---") && M("1--1--") && M("1---1-") && M("1----0") && M("-0-1--") && M("-0--1-") && M("-0---0") && M("--0-1-") && M("--0--0") && M("---1-0")) |
			// 100111:
			(M("10----") && M("-00---") && M("--01--") && M("---11-") && M("----11") && M("1-0---") && M("1--1--") && M("1---1-") && M("1----1") && M("-0-1--") && M("-0--1-") && M("-0---1") && M("--0-1-") && M("--0--1") && M("---1-1")) |

			// 101000:
			(M("10----") && M("-01---") && M("--10--") && M("---00-") && M("----00") && M("1-1---") && M("1--0--") && M("1---0-") && M("1----0") && M("-0-0--") && M("-0--0-") && M("-0---0") && M("--1-0-") && M("--1--0") && M("---0-0")) |
			// 101001:
			(M("10----") && M("-01---") && M("--10--") && M("---00-") && M("----01") && M("1-1---") && M("1--0--") && M("1---0-") && M("1----1") && M("-0-0--") && M("-0--0-") && M("-0---1") && M("--1-0-") && M("--1--1") && M("---0-1")) |
			// 101010:
			(M("10----") && M("-01---") && M("--10--") && M("---01-") && M("----10") && M("1-1---") && M("1--0--") && M("1---1-") && M("1----0") && M("-0-0--") && M("-0--1-") && M("-0---0") && M("--1-1-") && M("--1--0") && M("---0-0")) |
			// 101011:
			(M("10----") && M("-01---") && M("--10--") && M("---01-") && M("----11") && M("1-1---") && M("1--0--") && M("1---1-") && M("1----1") && M("-0-0--") && M("-0--1-") && M("-0---1") && M("--1-1-") && M("--1--1") && M("---0-1")) |
			// 101100:
			(M("10----") && M("-01---") && M("--11--") && M("---10-") && M("----00") && M("1-1---") && M("1--1--") && M("1---0-") && M("1----0") && M("-0-1--") && M("-0--0-") && M("-0---0") && M("--1-0-") && M("--1--0") && M("---1-0")) |
			// 101101:
			(M("10----") && M("-01---") && M("--11--") && M("---10-") && M("----01") && M("1-1---") && M("1--1--") && M("1---0-") && M("1----1") && M("-0-1--") && M("-0--0-") && M("-0---1") && M("--1-0-") && M("--1--1") && M("---1-1")) |
			// 101110:
			(M("10----") && M("-01---") && M("--11--") && M("---11-") && M("----10") && M("1-1---") && M("1--1--") && M("1---1-") && M("1----0") && M("-0-1--") && M("-0--1-") && M("-0---0") && M("--1-1-") && M("--1--0") && M("---1-0")) |
			// 101111:
			(M("10----") && M("-01---") && M("--11--") && M("---11-") && M("----11") && M("1-1---") && M("1--1--") && M("1---1-") && M("1----1") && M("-0-1--") && M("-0--1-") && M("-0---1") && M("--1-1-") && M("--1--1") && M("---1-1")) |

			// 110000:
			(M("11----") && M("-10---") && M("--00--") && M("---00-") && M("----00") && M("1-0---") && M("1--0--") && M("1---0-") && M("1----0") && M("-1-0--") && M("-1--0-") && M("-1---0") && M("--0-0-") && M("--0--0") && M("---0-0")) |
			// 110001:
			(M("11----") && M("-10---") && M("--00--") && M("---00-") && M("----01") && M("1-0---") && M("1--0--") && M("1---0-") && M("1----1") && M("-1-0--") && M("-1--0-") && M("-1---1") && M("--0-0-") && M("--0--1") && M("---0-1")) |
			// 110010:
			(M("11----") && M("-10---") && M("--00--") && M("---01-") && M("----10") && M("1-0---") && M("1--0--") && M("1---1-") && M("1----0") && M("-1-0--") && M("-1--1-") && M("-1---0") && M("--0-1-") && M("--0--0") && M("---0-0")) |
			// 110011:
			(M("11----") && M("-10---") && M("--00--") && M("---01-") && M("----11") && M("1-0---") && M("1--0--") && M("1---1-") && M("1----1") && M("-1-0--") && M("-1--1-") && M("-1---1") && M("--0-1-") && M("--0--1") && M("---0-1")) |
			// 110100:
			(M("11----") && M("-10---") && M("--01--") && M("---10-") && M("----00") && M("1-0---") && M("1--1--") && M("1---0-") && M("1----0") && M("-1-1--") && M("-1--0-") && M("-1---0") && M("--0-0-") && M("--0--0") && M("---1-0")) |
			// 110101:
			(M("11----") && M("-10---") && M("--01--") && M("---10-") && M("----01") && M("1-0---") && M("1--1--") && M("1---0-") && M("1----1") && M("-1-1--") && M("-1--0-") && M("-1---1") && M("--0-0-") && M("--0--1") && M("---1-1")) |
			// 110110:
			(M("11----") && M("-10---") && M("--01--") && M("---11-") && M("----10") && M("1-0---") && M("1--1--") && M("1---1-") && M("1----0") && M("-1-1--") && M("-1--1-") && M("-1---0") && M("--0-1-") && M("--0--0") && M("---1-0")) |
			// 110111:
			(M("11----") && M("-10---") && M("--01--") && M("---11-") && M("----11") && M("1-0---") && M("1--1--") && M("1---1-") && M("1----1") && M("-1-1--") && M("-1--1-") && M("-1---1") && M("--0-1-") && M("--0--1") && M("---1-1")) |

			// 111000:
			(M("11----") && M("-11---") && M("--10--") && M("---00-") && M("----00") && M("1-1---") && M("1--0--") && M("1---0-") && M("1----0") && M("-1-0--") && M("-1--0-") && M("-1---0") && M("--1-0-") && M("--1--0") && M("---0-0")) |
			// 111001:
			(M("11----") && M("-11---") && M("--10--") && M("---00-") && M("----01") && M("1-1---") && M("1--0--") && M("1---0-") && M("1----1") && M("-1-0--") && M("-1--0-") && M("-1---1") && M("--1-0-") && M("--1--1") && M("---0-1")) |
			// 111010:
			(M("11----") && M("-11---") && M("--10--") && M("---01-") && M("----10") && M("1-1---") && M("1--0--") && M("1---1-") && M("1----0") && M("-1-0--") && M("-1--1-") && M("-1---0") && M("--1-1-") && M("--1--0") && M("---0-0")) |
			// 111011:
			(M("11----") && M("-11---") && M("--10--") && M("---01-") && M("----11") && M("1-1---") && M("1--0--") && M("1---1-") && M("1----1") && M("-1-0--") && M("-1--1-") && M("-1---1") && M("--1-1-") && M("--1--1") && M("---0-1")) |
			// 111100:
			(M("11----") && M("-11---") && M("--11--") && M("---10-") && M("----00") && M("1-1---") && M("1--1--") && M("1---0-") && M("1----0") && M("-1-1--") && M("-1--0-") && M("-1---0") && M("--1-0-") && M("--1--0") && M("---1-0")) |
			// 111101:
			(M("11----") && M("-11---") && M("--11--") && M("---10-") && M("----01") && M("1-1---") && M("1--1--") && M("1---0-") && M("1----1") && M("-1-1--") && M("-1--0-") && M("-1---1") && M("--1-0-") && M("--1--1") && M("---1-1")) |
			// 111110:
			(M("11----") && M("-11---") && M("--11--") && M("---11-") && M("----10") && M("1-1---") && M("1--1--") && M("1---1-") && M("1----0") && M("-1-1--") && M("-1--1-") && M("-1---0") && M("--1-1-") && M("--1--0") && M("---1-0")) |
			// 111111:
			(M("11----") && M("-11---") && M("--11--") && M("---11-") && M("----11") && M("1-1---") && M("1--1--") && M("1---1-") && M("1----1") && M("-1-1--") && M("-1--1-") && M("-1---1") && M("--1-1-") && M("--1--1") && M("---1-1"));
	}

	printf("should not get here!\n");
	return false;
}

bool IsInAnyStillUsedCTLine(int ClauseIndex) // for debugging only
{
	// this function returns true if PossibleClauses[ClauseIndex]
	// is in any ClauseTable line that is still 'in use', that means
	// whose clauses at left of ClauseIndex are ALL absent in the CNF.

	for (int i = 0; i < CTLineNumber; i ++) // loop through the ClauseTable lines
	{
		for (int j = 0; j < CTColumnNumber; j ++) // loop through the ClauseTable columns
		{
			if (ExistsInCNF(CT[i][j]))
				break; // goto next ClauseTable line

			if (IsEqual(CT[i][j], PossibleClauses[ClauseIndex]))
				return true;
		}
	}

	return false;
}

// *** <<< END OF NOT REQUIRED FOR ACTUAL SOLVER BUT FOR DEBUGGING/VISUALIZATION ONLY ***



// **************************************************************

bool Does_CNF_Have_A_Solution() // uses some global variables defined at top of this code file
{
	// some preparation...

	if (DigitNumberOld != DigitNumber ||
		SAT_TYPE_OLD != SAT_TYPE) // at first run or when DigitNumber or/and SAT_TYPE changed
	{
		DigitNumberOld = DigitNumber;
		SAT_TYPE_OLD = SAT_TYPE;

		if (SAT_TYPE == 2)
		{
			CombinationCount = 4; // 00, 01, 10, 11
		}
		if (SAT_TYPE == 3)
		{
			CombinationCount = 8; // 000, 001, 010, 011, 100, 101, 110, 111
		}

		ComplexityMax = 0; // reset maximal complexity

		PossibleClauses_Create(); // initialize PossibleClauses[][] and PossibleClauseNumber

		if (DigitNumber <= CT_DIGIT_NUMBER_MAX) // the ClauseTable is not initialized for too large DigitNumbers!
		{
			ClauseTable_CreateDynamically(); // used for debugging ONLY as exponential memory/CPU usage!
		}
	}

	// POLYNOMIAL SOLVER >>>

	/*
	// for debugging
	if (DigitNumber <= CT_DIGIT_NUMBER_MAX)
		ClauseTable_Show();
	*/

	ExistsInCNF_Cache();

	// TREAT SPECIAL CNFS

	// NOTE: if the ClauseTable has less than 3 columns, the polynomial
	// algorithm won't work. But this is only the case for trivial CNFs
	// that have no practical meaning, like the following ones.
	// Therefore we implement a special treatment.

	if (SAT_TYPE == 2 && DigitNumber == 2)
		return (
			M("00") ||
			M("01") ||
			M("10") ||
			M("11")
			);

	if (SAT_TYPE == 3 && DigitNumber == 3)
		return (
			M("000") ||
			M("001") ||
			M("010") ||
			M("011") ||
			M("100") ||
			M("101") ||
			M("110") ||
			M("111")
			);

	// END OF TREATING SPECIAL CNFS

	static bool StillToBeChecked[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX];
	static bool StillToBeChecked_New[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX];

	// initialize with true
	for (int i = 0; i < PossibleClauseNumber; i ++)
		for (int j = 0; j < PossibleClauseNumber; j ++)
			StillToBeChecked[i][j] = true;

	// Imagine the ClauseTable was:
	// 
	// (00- & 0-0 & -00) |   <-'ClauseTable line'
	// (00- & 0-1 & -01) |
	// (01- & 0-0 & -10) |
	// (01- & 0-1 & -11) |
	// (10- & 1-0 & -00) |
	// (10- & 1-1 & -01) |
	// (11- & 1-0 & -10) |
	// (11- & 1-1 & -11)
	//  ^'ClauseTable column'
	// 
	// (actually a trivial example, but I use it here because it's easy to understand.)

	for (int i = 0; i < PossibleClauseNumber; i ++) // loops through all thinkable clauses
	{
		Complexity1 ++; // to determine complexity; not involved in actual solving

		if ((i % CombinationCount) == 0) // when i points to first clause of a ClauseTable column
		{
			// we go here for i = 00-, 0-0, -00
			for (int m = 0; m < PossibleClauseNumber; m ++)
				for (int j = 0; j < PossibleClauseNumber; j ++)
					StillToBeChecked_New[m][j] = false; // initialize/reset
		}

		for (int j = i + 1; j < PossibleClauseNumber; j ++) // begin at next PossibleClause behind PossibleClause[i]
		{
			Complexity2 ++;

			if (StillToBeChecked[i][j] == false)
				continue; // next j iteration

			if (ExistsInCNF(i) == false &&
				ExistsInCNF(j) == false &&
				IsInConflict(PossibleClauses[i], PossibleClauses[j]) == false)
			{
				if (i < PossibleClauseNumber - CombinationCount * 2) // are there at least two ClauseTable columns 'behind' i?
				{
					for (int k = j + 1; k < PossibleClauseNumber; k ++) // begin at next PossibleClause behind PossibleClause[j]
					{
						Complexity3 ++;

						if (StillToBeChecked[i][k] == false)
							continue; // next k iteration

						if (ExistsInCNF(i) == false &&
							ExistsInCNF(k) == false &&
							IsInConflict(PossibleClauses[i], PossibleClauses[k]) == false &&
							IsInConflict(PossibleClauses[j], PossibleClauses[k]) == false)
						{
							StillToBeChecked_New[j][k] = true;
						}
					}
				}
				else
				{
					// i, j and k reached their 'right-most position' in the ClauseTable
					if (StillToBeChecked[i][j])
						return true; // CNF has any solution
				}
			}
		}

		if ((i % CombinationCount) == (CombinationCount - 1)) // when i points to the last clause of a ClauseTable column
		{
			// we go here for i = 11-, 1-1, -11
			for (int m = 0; m < PossibleClauseNumber; m ++)
				for (int j = 0; j < PossibleClauseNumber; j ++)
					StillToBeChecked[m][j] &= StillToBeChecked_New[m][j]; // StillToBeChecked stays true if it was true before and also StillToBeChecked_New is true
		}
	}

	return false; // CNF has no solution

	// <<< END OF POLYNOMIAL SOLVER
}

// **************************************************************

void ShowComplexity(char* InfoText, u64 Complexity)
{
	printf(InfoText);

	if (Complexity > 1000000000)
		// the following should work, tested
		printf("%d%d\n", (int)(Complexity / 1000000000I64), (int)(Complexity % 1000000000I64));
	else
		printf("%d\n", (int)(Complexity));
}

bool IsCNFValid(int DigitNumber, int ClauseNumber, char* CNF, int SAT_TYPE)
{
	// This is used to make sure the polynomial solver is
	// used in the way it's meant to be used.
	// This function here is NOT required for the actual solving!

	switch (SAT_TYPE)
	{
	case 2:

		for (int i = 0; i < ClauseNumber; i ++)
		{
			int UsedDigitNumber = 0;

			for (int j = 0; j < DigitNumber; j ++)
				if (CNF[i * DigitNumber + j] != '-')
					UsedDigitNumber ++;

			if (UsedDigitNumber != 2) // 2
				return false;
		}

		return true;

	case 3:

		for (int i = 0; i < ClauseNumber; i ++)
		{
			int UsedDigitNumber = 0;

			for (int j = 0; j < DigitNumber; j ++)
				if (CNF[i * DigitNumber + j] != '-')
					UsedDigitNumber ++;

			if (UsedDigitNumber != 3) // 3
				return false;
		}

		return true;

	default:

		return false;
	}
}

bool CNF_Solve_Polynomial(int DigitNumberPassed, int ClauseNumberPassed, char* CNFPassed, bool& SolvingAbortedFlag) // this is the one and only function to be used by the target project (i.e. here the procedure 'main()')
{ // this is the main procedure of the polynomial solver; it returns true if the passed CNF is solvable, false if not

	// verify
	if (IsCNFValid(DigitNumberPassed, ClauseNumberPassed, CNFPassed, SAT_TYPE) == false)
	{
		printf("ERROR: something's wrong with the CNF to solve.\n");
		printf("Maybe you tried to solve a 3-SAT CNF in 2-SAT mode\n");
		printf("(or vice versa) or the CNF is neither 2-SAT nor 3-SAT.\n");
		printf("The polynomial solving can NOT be done!\n");
		printf("\n");
		printf("Notice that you can press 's' in the main menu to\n");
		printf("switch between the solver's 2-SAT or 3-SAT mode!\n");
		getchar();

		SolvingAbortedFlag = true;
		return false;
	}

	// set the global vars
	DigitNumber = DigitNumberPassed;
	ClauseNumber = ClauseNumberPassed;
	CNF = CNFPassed;

	// Example usage:
	// bool CNFSolvable = CNF_Solve_Polynomial(
	//  4,
	//  8,
	//  "000-"   [NO comma!]
	//  "001-"
	//  "010-"
	//  "011-"
	//  "100-"
	//  "101-"
	//  "110-"
	//  "111-");
	// 
	// [CNFSolvable = false].

	// reset the (current) complexity counters
	Complexity1 = 0;
	Complexity2 = 0;
	Complexity3 = 0;
	Complexity4 = 0;

	bool Solvable = Does_CNF_Have_A_Solution(); // the actual polynomial solver procedure

#ifndef MIN_OUTPUT
	// some debug output
	/*
	if (Solvable)
		printf("Polynomial: Satisfiable\n");
	else
		printf("Polynomial: UNSAT\n");
	*/
#endif

	// determine the highest complexity ever appeared
	if (ComplexityMax < Complexity1)
		ComplexityMax = Complexity1;
	if (ComplexityMax < Complexity2)
		ComplexityMax = Complexity2;
	if (ComplexityMax < Complexity3)
		ComplexityMax = Complexity3;
	if (ComplexityMax < Complexity4)
		ComplexityMax = Complexity4;
	
	// NOTE: the following expected complexity calculation may differ from the one
	// shown in the Algorithm Explanation document. This is because here are some
	// implementation details regarded that are ignored in the Algorithm Explanation.

	u64 ComplexityMaxExpected =
		// 
		// NOTE: the complexity is calculated as the iteration count of the three nested loops
		// i, j, k in the main part of the polynomial solver and the calls to IsInConflict().
		// That following calculation is not very exact but should be an upper bound that
		// is never exceeded.
		// 
		(u64)PossibleClauseNumber * (u64)(PossibleClauseNumber / 2) * (u64)(PossibleClauseNumber / 4) * (u64)(3 * DigitNumber);

#ifndef MIN_OUTPUT
	// show complexity information
	printf("   ************ Polynomial Complexity Info ************\n");
	ShowComplexity("   * Complexity1=", Complexity1);
	ShowComplexity("   * Complexity2=", Complexity2);
	ShowComplexity("   * Complexity3=", Complexity3);
	ShowComplexity("   * Complexity4=", Complexity4);
	ShowComplexity("   * ComplexityMaxExpected=", ComplexityMaxExpected);
	ShowComplexity("   * ComplexityMax=", ComplexityMax);
	printf("   * P. solver used %d%% of expected maximal complexity\n", ComplexityMax * 100I64 / ComplexityMaxExpected);
	printf("   ****************************************************\n");
#endif

	// CLAUSE TABLE TEST >>>

	if (SAT_TYPE == 2 && DigitNumber >= 4 && DigitNumber <= 6)
	{
		bool ClauseTableSolved = (ClauseTableTest() != 0); // calls the function 'ClauseTableTest()'
		if (ClauseTableSolved != Solvable)
		{
			printf("ERROR: hard-coded, exponential ClauseTable solver returned\n");
			printf("       something different from the polynomial solver!\n");
			printf("       This should not happen, something's wrong!\n");

			getchar();
		}
		else
		{
			// if we go here, all's alright!
			// printf("hard-coded, exponential ClauseTable solver result matches polynomial one.\n");
			// printf("\n");
		}
	}

	// <<< END IF CLAUSE TABLE TEST

	// SolvingAbortedFlag = true; // enable to NOT check if polynomial solver returned the right result

	return Solvable;
}

// *******************************************
// *** <<< END OF ACTUAL POLYNOMIAL SOLVER ***
// *******************************************
