// (c)2015-2018 by Louis Coder. Additional debugging/testing code of Polynomial SAT Solver vD2-1.1.

// ********** NOT REQUIRED FOR ACTUAL SOLVER BUT FOR DEBUGGING ONLY ***********

// PLEASE NOTE: the following code is only intended for additional testing/debugging
// of the actual Polynomial Solver. You could remove all code within this block
// ("NOT REQUIRED FOR ACTUAL SOLVER BUT..."), and the Polynomial Solver would still
// work! In particular, ClauseTable_CreateDynamically() needs exponential time and
// space, but is ONLY implemented as a debugging help (which I required during
// development of the polynomial SAT solving algorithm).

int GetPossibleClauseIndex(const char* PossibleClause) // rather slow - use seldom only!
{
	for (int i = 0; i < PossibleClauseNumber; i ++)
		if (PossibleClause[0] == PossibleClauses[i][0] &&
			PossibleClause[1] == PossibleClauses[i][1]) // quick incomplete check to increase speed
			if (IsEqual(PossibleClause, PossibleClauses[i]))
				return i;

	printf("ERROR in GetPossibleClauseIndex(): argument is not a PossibleClause or ClauseTable has not been created correctly !\n");
	getchar();
	return -1;
}

void ClauseTable_CreateDynamically() // fills CT[][][] and sets CTLineNumber and CTColumnNumber
{ // PossibleClauses_Create() must have been called before using this procedure!

	// IMPORTANT: this procedures needs O(2^n) much time and space,
	// but it is NOT required for the actual polynomial solving, but
	// only for debugging/optional error checking!!!

	CTLineNumber = 0;

	if (DigitNumber > CT_DIGIT_NUMBER_MAX)
	{
		printf("ERROR: ClauseTable_CreateDynamically() can only\n");
		printf("       be used for a DigitNumber <= %d !\n", CT_DIGIT_NUMBER_MAX);
		printf("       You tried to use it for DigitNumber = %d.\n", DigitNumber);
		getchar();
	}

	char CurrentLine[DIGIT_NUMBER_MAX + 1]; // '+ 1' for an eventual null terminator

	for (int i = 0; i < DigitNumber; i ++)
		CurrentLine[i] = '0';

	// NOTE: CurrentLine is initially (for DigitNumber = 5):
	// 00000.
	// In next while-iteration, it is:
	// 00001.
	// Then
	// 00010,
	// 00011,
	// 00100,
	// 00101,
	// 00110,
	// 00111, ...
	// Notice that this is EXPONENTIAL, but used for debugging/visualization ONLY!

	while (true)
	{
		for (int i = 0; i < CT_COLUMN_NUMBER_MAX; i ++)
			for (int m = 0; m < CT_DIGIT_NUMBER_MAX; m ++)
				CT[CTLineNumber][i][m] = '\0'; // clear

		int CTColumnIndex = 0;

		if (SAT_TYPE == 2)
		{
			for (int D1 = 0; D1 < DigitNumber - 1; D1 ++)
			{
				for (int D2 = D1 + 1; D2 < DigitNumber; D2 ++)
				{
					for (int m = 0; m < DigitNumber; m ++)
						CT[CTLineNumber][CTColumnIndex][m] = '-';

					CT[CTLineNumber][CTColumnIndex][D1] = CurrentLine[D1];
					CT[CTLineNumber][CTColumnIndex][D2] = CurrentLine[D2];
					CT[CTLineNumber][CTColumnIndex][DigitNumber] = '\0';
					CTIndex[CTLineNumber][CTColumnIndex] = GetPossibleClauseIndex(CT[CTLineNumber][CTColumnIndex]);

					if (CTColumnIndex >= CT_COLUMN_NUMBER_MAX)
					{
						printf("CTColumnIndex overflow !");
						getchar();
					}

					CTColumnIndex ++;
				}
			}
		}

		if (SAT_TYPE == 3)
		{
			for (int D1 = 0; D1 < DigitNumber - 2; D1 ++)
			{
				for (int D2 = D1 + 1; D2 < DigitNumber - 1; D2 ++)
				{
					for (int D3 = D2 + 1; D3 < DigitNumber; D3 ++)
					{
						for (int m = 0; m < DigitNumber; m ++)
							CT[CTLineNumber][CTColumnIndex][m] = '-';

						CT[CTLineNumber][CTColumnIndex][D1] = CurrentLine[D1];
						CT[CTLineNumber][CTColumnIndex][D2] = CurrentLine[D2];
						CT[CTLineNumber][CTColumnIndex][D3] = CurrentLine[D3];
						CT[CTLineNumber][CTColumnIndex][DigitNumber] = '\0';
						CTIndex[CTLineNumber][CTColumnIndex] = GetPossibleClauseIndex(CT[CTLineNumber][CTColumnIndex]);

						if (CTColumnIndex >= CT_COLUMN_NUMBER_MAX)
						{
							printf("CTColumnIndex overflow !");
							getchar();
						}

						CTColumnIndex ++;
					}
				}
			}
		}

		if (CTLineNumber >= CT_LINE_NUMBER_MAX)
		{
			printf("CTLineNumber overflow !");
			getchar();
		}

		CTLineNumber ++;
		CTColumnNumber = CTColumnIndex;

		for (int m = DigitNumber - 1; m >= 0; m --)
			if (CurrentLine[m] == '0')
			{
				CurrentLine[m] = '1';
				goto NextLine;
			}
			else if (CurrentLine[m] == '1')
				CurrentLine[m] = '0';

		break;
NextLine:;
	}
}

void ClauseTable_Show()
{ // shows the whole ClauseTable in the Windows console or in a Linux shell

	printf("CLAUSE TABLE >>>\n\n");

	for (int i = 0; i < CTLineNumber; i ++) // loop through the ClauseTable lines
	{
		// show header that tells which PossibleClauseIndices the clauses have
		if (i == 0)
		{
			printf("['C (X)' means clause C has PossibleClauseIndex X. '!' means clause is false as not in SAT CNF.]\n\n");

			for (int j = 0; j < CTColumnNumber; j ++) // loop through the ClauseTable columns
			{
				static char S[1024 * 1024];
				S[0] = '\0'; // important for printf()

				SPRINTF2(S, " %d to %d...", (j + 0) * CombinationCount, (j + 1) * CombinationCount - 1);

				printf("%s", S);

				for (int m = DigitNumber + 6 + 2 - (int)strlen(S); m >= 0; m --)
					printf(" ");
			}
			printf("\n\n");
		}

		// show the clauses and their PossibleClauses[][]-indices
		for (int j = 0; j < CTColumnNumber; j ++)
		{
			int PossibleClauseIndex = GetPossibleClauseIndex(CT[i][j]);

			if (ExistsInCNF(PossibleClauseIndex))
				printf("!"); // clause is thought to be 'false' in the mathematic model of the ClauseTable
			else
				printf(" "); // clause is thought to be 'true'...

			for (int m = 0; m < DigitNumber; m ++)
				printf("%c", CT[i][j][m]);

			if (PossibleClauseIndex < 10)
				printf(" (%d)  ", PossibleClauseIndex);
			else
			if (PossibleClauseIndex < 100)
				printf(" (%d) ", PossibleClauseIndex);
			else
				printf(" (%d)", PossibleClauseIndex);

			for (int m = 1; m >= 0; m --)
				printf(" ");
		}

		printf("\n");
	}

	printf("\n<<< END OF CLAUSE TABLE\n\n");
}

bool M(const char* Clause) // [clause] Missing in CNF to solve?
{
	return !ExistsInCNF(Clause);
}

int ClauseTableTest()
{
	// This procedure shall show that you can really determine if a CNF is solvable
	// or not by using the 'ClauseTable'. The following implementation of the
	// 'ClauseTable' works for 2-SAT CNFs with a DigitNumber of 4, 5 or 6.
	// 
	// You could construct a 'ClauseTable' for 3-SAT as well, there's no significant
	// difference, except that the size of the table (number of M("...") calls) is
	// different.

	// IMPORTANT NOTE: this procedure solves SAT CNFs with an exponential amount of
	// time and space. That's why it is used for demonstration only, the actual
	// Polynomial Solver does NOT require/use ClauseTableTest() in any way!!!

	switch (DigitNumber)
	{
	case 4: // 2^4 = 16 ClauseTableLines, 6 ClauseTableColumns
		return
			(M("00--") && M("-00-") && M("--00") && M("0-0-") && M("-0-0") && M("0--0")) |
			(M("00--") && M("-00-") && M("--01") && M("0-0-") && M("-0-1") && M("0--1")) |
			(M("00--") && M("-01-") && M("--10") && M("0-1-") && M("-0-0") && M("0--0")) |
			(M("00--") && M("-01-") && M("--11") && M("0-1-") && M("-0-1") && M("0--1")) |
			(M("01--") && M("-10-") && M("--00") && M("0-0-") && M("-1-0") && M("0--0")) |
			(M("01--") && M("-10-") && M("--01") && M("0-0-") && M("-1-1") && M("0--1")) |
			(M("01--") && M("-11-") && M("--10") && M("0-1-") && M("-1-0") && M("0--0")) |
			(M("01--") && M("-11-") && M("--11") && M("0-1-") && M("-1-1") && M("0--1")) |
			(M("10--") && M("-00-") && M("--00") && M("1-0-") && M("-0-0") && M("1--0")) |
			(M("10--") && M("-00-") && M("--01") && M("1-0-") && M("-0-1") && M("1--1")) |
			(M("10--") && M("-01-") && M("--10") && M("1-1-") && M("-0-0") && M("1--0")) |
			(M("10--") && M("-01-") && M("--11") && M("1-1-") && M("-0-1") && M("1--1")) |
			(M("11--") && M("-10-") && M("--00") && M("1-0-") && M("-1-0") && M("1--0")) |
			(M("11--") && M("-10-") && M("--01") && M("1-0-") && M("-1-1") && M("1--1")) |
			(M("11--") && M("-11-") && M("--10") && M("1-1-") && M("-1-0") && M("1--0")) |
			(M("11--") && M("-11-") && M("--11") && M("1-1-") && M("-1-1") && M("1--1"));

	case 5: // 2^5 = 32 ClauseTableLines, 10 ClauseTableColumns
		return
			// 00000:
			(M("00---") && M("-00--") && M("--00-") && M("---00") && M("0-0--") && M("0--0-") && M("0---0") && M("-0-0-") && M("-0--0") && M("--0-0")) |
			// 00001:
			(M("00---") && M("-00--") && M("--00-") && M("---01") && M("0-0--") && M("0--0-") && M("0---1") && M("-0-0-") && M("-0--1") && M("--0-1")) |
			// 00010:
			(M("00---") && M("-00--") && M("--01-") && M("---10") && M("0-0--") && M("0--1-") && M("0---0") && M("-0-1-") && M("-0--0") && M("--0-0")) |
			// 00011:
			(M("00---") && M("-00--") && M("--01-") && M("---11") && M("0-0--") && M("0--1-") && M("0---1") && M("-0-1-") && M("-0--1") && M("--0-1")) |
			// 00100:
			(M("00---") && M("-01--") && M("--10-") && M("---00") && M("0-1--") && M("0--0-") && M("0---0") && M("-0-0-") && M("-0--0") && M("--1-0")) |
			// 00101:
			(M("00---") && M("-01--") && M("--10-") && M("---01") && M("0-1--") && M("0--0-") && M("0---1") && M("-0-0-") && M("-0--1") && M("--1-1")) |
			// 00110:
			(M("00---") && M("-01--") && M("--11-") && M("---10") && M("0-1--") && M("0--1-") && M("0---0") && M("-0-1-") && M("-0--0") && M("--1-0")) |
			// 00111:
			(M("00---") && M("-01--") && M("--11-") && M("---11") && M("0-1--") && M("0--1-") && M("0---1") && M("-0-1-") && M("-0--1") && M("--1-1")) |

			// 01000:
			(M("01---") && M("-10--") && M("--00-") && M("---00") && M("0-0--") && M("0--0-") && M("0---0") && M("-1-0-") && M("-1--0") && M("--0-0")) |
			// 01001:
			(M("01---") && M("-10--") && M("--00-") && M("---01") && M("0-0--") && M("0--0-") && M("0---1") && M("-1-0-") && M("-1--1") && M("--0-1")) |
			// 01010:
			(M("01---") && M("-10--") && M("--01-") && M("---10") && M("0-0--") && M("0--1-") && M("0---0") && M("-1-1-") && M("-1--0") && M("--0-0")) |
			// 01011:
			(M("01---") && M("-10--") && M("--01-") && M("---11") && M("0-0--") && M("0--1-") && M("0---1") && M("-1-1-") && M("-1--1") && M("--0-1")) |
			// 01100:
			(M("01---") && M("-11--") && M("--10-") && M("---00") && M("0-1--") && M("0--0-") && M("0---0") && M("-1-0-") && M("-1--0") && M("--1-0")) |
			// 01101:
			(M("01---") && M("-11--") && M("--10-") && M("---01") && M("0-1--") && M("0--0-") && M("0---1") && M("-1-0-") && M("-1--1") && M("--1-1")) |
			// 01110:
			(M("01---") && M("-11--") && M("--11-") && M("---10") && M("0-1--") && M("0--1-") && M("0---0") && M("-1-1-") && M("-1--0") && M("--1-0")) |
			// 01111:
			(M("01---") && M("-11--") && M("--11-") && M("---11") && M("0-1--") && M("0--1-") && M("0---1") && M("-1-1-") && M("-1--1") && M("--1-1")) |

			// 10000:
			(M("10---") && M("-00--") && M("--00-") && M("---00") && M("1-0--") && M("1--0-") && M("1---0") && M("-0-0-") && M("-0--0") && M("--0-0")) |
			// 10001:
			(M("10---") && M("-00--") && M("--00-") && M("---01") && M("1-0--") && M("1--0-") && M("1---1") && M("-0-0-") && M("-0--1") && M("--0-1")) |
			// 10010:
			(M("10---") && M("-00--") && M("--01-") && M("---10") && M("1-0--") && M("1--1-") && M("1---0") && M("-0-1-") && M("-0--0") && M("--0-0")) |
			// 10011:
			(M("10---") && M("-00--") && M("--01-") && M("---11") && M("1-0--") && M("1--1-") && M("1---1") && M("-0-1-") && M("-0--1") && M("--0-1")) |
			// 10100:
			(M("10---") && M("-01--") && M("--10-") && M("---00") && M("1-1--") && M("1--0-") && M("1---0") && M("-0-0-") && M("-0--0") && M("--1-0")) |
			// 10101:
			(M("10---") && M("-01--") && M("--10-") && M("---01") && M("1-1--") && M("1--0-") && M("1---1") && M("-0-0-") && M("-0--1") && M("--1-1")) |
			// 10110:
			(M("10---") && M("-01--") && M("--11-") && M("---10") && M("1-1--") && M("1--1-") && M("1---0") && M("-0-1-") && M("-0--0") && M("--1-0")) |
			// 10111:
			(M("10---") && M("-01--") && M("--11-") && M("---11") && M("1-1--") && M("1--1-") && M("1---1") && M("-0-1-") && M("-0--1") && M("--1-1")) |

			// 11000:
			(M("11---") && M("-10--") && M("--00-") && M("---00") && M("1-0--") && M("1--0-") && M("1---0") && M("-1-0-") && M("-1--0") && M("--0-0")) |
			// 11001:
			(M("11---") && M("-10--") && M("--00-") && M("---01") && M("1-0--") && M("1--0-") && M("1---1") && M("-1-0-") && M("-1--1") && M("--0-1")) |
			// 11010:
			(M("11---") && M("-10--") && M("--01-") && M("---10") && M("1-0--") && M("1--1-") && M("1---0") && M("-1-1-") && M("-1--0") && M("--0-0")) |
			// 11011:
			(M("11---") && M("-10--") && M("--01-") && M("---11") && M("1-0--") && M("1--1-") && M("1---1") && M("-1-1-") && M("-1--1") && M("--0-1")) |
			// 11100:
			(M("11---") && M("-11--") && M("--10-") && M("---00") && M("1-1--") && M("1--0-") && M("1---0") && M("-1-0-") && M("-1--0") && M("--1-0")) |
			// 11101:
			(M("11---") && M("-11--") && M("--10-") && M("---01") && M("1-1--") && M("1--0-") && M("1---1") && M("-1-0-") && M("-1--1") && M("--1-1")) |
			// 11110:
			(M("11---") && M("-11--") && M("--11-") && M("---10") && M("1-1--") && M("1--1-") && M("1---0") && M("-1-1-") && M("-1--0") && M("--1-0")) |
			// 11111:
			(M("11---") && M("-11--") && M("--11-") && M("---11") && M("1-1--") && M("1--1-") && M("1---1") && M("-1-1-") && M("-1--1") && M("--1-1"));

	case 6: // 2^6 = 64 ClauseTableLines, 15 ClauseTableColumns
		return
			// 000000:
			(M("00----") && M("-00---") && M("--00--") && M("---00-") && M("----00") && M("0-0---") && M("0--0--") && M("0---0-") && M("0----0") && M("-0-0--") && M("-0--0-") && M("-0---0") && M("--0-0-") && M("--0--0") && M("---0-0")) |
			// 000001:
			(M("00----") && M("-00---") && M("--00--") && M("---00-") && M("----01") && M("0-0---") && M("0--0--") && M("0---0-") && M("0----1") && M("-0-0--") && M("-0--0-") && M("-0---1") && M("--0-0-") && M("--0--1") && M("---0-1")) |
			// 000010:
			(M("00----") && M("-00---") && M("--00--") && M("---01-") && M("----10") && M("0-0---") && M("0--0--") && M("0---1-") && M("0----0") && M("-0-0--") && M("-0--1-") && M("-0---0") && M("--0-1-") && M("--0--0") && M("---0-0")) |
			// 000011:
			(M("00----") && M("-00---") && M("--00--") && M("---01-") && M("----11") && M("0-0---") && M("0--0--") && M("0---1-") && M("0----1") && M("-0-0--") && M("-0--1-") && M("-0---1") && M("--0-1-") && M("--0--1") && M("---0-1")) |
			// 000100:
			(M("00----") && M("-00---") && M("--01--") && M("---10-") && M("----00") && M("0-0---") && M("0--1--") && M("0---0-") && M("0----0") && M("-0-1--") && M("-0--0-") && M("-0---0") && M("--0-0-") && M("--0--0") && M("---1-0")) |
			// 000101:
			(M("00----") && M("-00---") && M("--01--") && M("---10-") && M("----01") && M("0-0---") && M("0--1--") && M("0---0-") && M("0----1") && M("-0-1--") && M("-0--0-") && M("-0---1") && M("--0-0-") && M("--0--1") && M("---1-1")) |
			// 000110:
			(M("00----") && M("-00---") && M("--01--") && M("---11-") && M("----10") && M("0-0---") && M("0--1--") && M("0---1-") && M("0----0") && M("-0-1--") && M("-0--1-") && M("-0---0") && M("--0-1-") && M("--0--0") && M("---1-0")) |
			// 000111:
			(M("00----") && M("-00---") && M("--01--") && M("---11-") && M("----11") && M("0-0---") && M("0--1--") && M("0---1-") && M("0----1") && M("-0-1--") && M("-0--1-") && M("-0---1") && M("--0-1-") && M("--0--1") && M("---1-1")) |

			// 001000:
			(M("00----") && M("-01---") && M("--10--") && M("---00-") && M("----00") && M("0-1---") && M("0--0--") && M("0---0-") && M("0----0") && M("-0-0--") && M("-0--0-") && M("-0---0") && M("--1-0-") && M("--1--0") && M("---0-0")) |
			// 001001:
			(M("00----") && M("-01---") && M("--10--") && M("---00-") && M("----01") && M("0-1---") && M("0--0--") && M("0---0-") && M("0----1") && M("-0-0--") && M("-0--0-") && M("-0---1") && M("--1-0-") && M("--1--1") && M("---0-1")) |
			// 001010:
			(M("00----") && M("-01---") && M("--10--") && M("---01-") && M("----10") && M("0-1---") && M("0--0--") && M("0---1-") && M("0----0") && M("-0-0--") && M("-0--1-") && M("-0---0") && M("--1-1-") && M("--1--0") && M("---0-0")) |
			// 001011:
			(M("00----") && M("-01---") && M("--10--") && M("---01-") && M("----11") && M("0-1---") && M("0--0--") && M("0---1-") && M("0----1") && M("-0-0--") && M("-0--1-") && M("-0---1") && M("--1-1-") && M("--1--1") && M("---0-1")) |
			// 001100:
			(M("00----") && M("-01---") && M("--11--") && M("---10-") && M("----00") && M("0-1---") && M("0--1--") && M("0---0-") && M("0----0") && M("-0-1--") && M("-0--0-") && M("-0---0") && M("--1-0-") && M("--1--0") && M("---1-0")) |
			// 001101:
			(M("00----") && M("-01---") && M("--11--") && M("---10-") && M("----01") && M("0-1---") && M("0--1--") && M("0---0-") && M("0----1") && M("-0-1--") && M("-0--0-") && M("-0---1") && M("--1-0-") && M("--1--1") && M("---1-1")) |
			// 001110:
			(M("00----") && M("-01---") && M("--11--") && M("---11-") && M("----10") && M("0-1---") && M("0--1--") && M("0---1-") && M("0----0") && M("-0-1--") && M("-0--1-") && M("-0---0") && M("--1-1-") && M("--1--0") && M("---1-0")) |
			// 001111:
			(M("00----") && M("-01---") && M("--11--") && M("---11-") && M("----11") && M("0-1---") && M("0--1--") && M("0---1-") && M("0----1") && M("-0-1--") && M("-0--1-") && M("-0---1") && M("--1-1-") && M("--1--1") && M("---1-1")) |

			// 010000:
			(M("01----") && M("-10---") && M("--00--") && M("---00-") && M("----00") && M("0-0---") && M("0--0--") && M("0---0-") && M("0----0") && M("-1-0--") && M("-1--0-") && M("-1---0") && M("--0-0-") && M("--0--0") && M("---0-0")) |
			// 010001:
			(M("01----") && M("-10---") && M("--00--") && M("---00-") && M("----01") && M("0-0---") && M("0--0--") && M("0---0-") && M("0----1") && M("-1-0--") && M("-1--0-") && M("-1---1") && M("--0-0-") && M("--0--1") && M("---0-1")) |
			// 010010:
			(M("01----") && M("-10---") && M("--00--") && M("---01-") && M("----10") && M("0-0---") && M("0--0--") && M("0---1-") && M("0----0") && M("-1-0--") && M("-1--1-") && M("-1---0") && M("--0-1-") && M("--0--0") && M("---0-0")) |
			// 010011:
			(M("01----") && M("-10---") && M("--00--") && M("---01-") && M("----11") && M("0-0---") && M("0--0--") && M("0---1-") && M("0----1") && M("-1-0--") && M("-1--1-") && M("-1---1") && M("--0-1-") && M("--0--1") && M("---0-1")) |
			// 010100:
			(M("01----") && M("-10---") && M("--01--") && M("---10-") && M("----00") && M("0-0---") && M("0--1--") && M("0---0-") && M("0----0") && M("-1-1--") && M("-1--0-") && M("-1---0") && M("--0-0-") && M("--0--0") && M("---1-0")) |
			// 010101:
			(M("01----") && M("-10---") && M("--01--") && M("---10-") && M("----01") && M("0-0---") && M("0--1--") && M("0---0-") && M("0----1") && M("-1-1--") && M("-1--0-") && M("-1---1") && M("--0-0-") && M("--0--1") && M("---1-1")) |
			// 010110:
			(M("01----") && M("-10---") && M("--01--") && M("---11-") && M("----10") && M("0-0---") && M("0--1--") && M("0---1-") && M("0----0") && M("-1-1--") && M("-1--1-") && M("-1---0") && M("--0-1-") && M("--0--0") && M("---1-0")) |
			// 010111:
			(M("01----") && M("-10---") && M("--01--") && M("---11-") && M("----11") && M("0-0---") && M("0--1--") && M("0---1-") && M("0----1") && M("-1-1--") && M("-1--1-") && M("-1---1") && M("--0-1-") && M("--0--1") && M("---1-1")) |

			// 011000:
			(M("01----") && M("-11---") && M("--10--") && M("---00-") && M("----00") && M("0-1---") && M("0--0--") && M("0---0-") && M("0----0") && M("-1-0--") && M("-1--0-") && M("-1---0") && M("--1-0-") && M("--1--0") && M("---0-0")) |
			// 011001:
			(M("01----") && M("-11---") && M("--10--") && M("---00-") && M("----01") && M("0-1---") && M("0--0--") && M("0---0-") && M("0----1") && M("-1-0--") && M("-1--0-") && M("-1---1") && M("--1-0-") && M("--1--1") && M("---0-1")) |
			// 011010:
			(M("01----") && M("-11---") && M("--10--") && M("---01-") && M("----10") && M("0-1---") && M("0--0--") && M("0---1-") && M("0----0") && M("-1-0--") && M("-1--1-") && M("-1---0") && M("--1-1-") && M("--1--0") && M("---0-0")) |
			// 011011:
			(M("01----") && M("-11---") && M("--10--") && M("---01-") && M("----11") && M("0-1---") && M("0--0--") && M("0---1-") && M("0----1") && M("-1-0--") && M("-1--1-") && M("-1---1") && M("--1-1-") && M("--1--1") && M("---0-1")) |
			// 011100:
			(M("01----") && M("-11---") && M("--11--") && M("---10-") && M("----00") && M("0-1---") && M("0--1--") && M("0---0-") && M("0----0") && M("-1-1--") && M("-1--0-") && M("-1---0") && M("--1-0-") && M("--1--0") && M("---1-0")) |
			// 011101:
			(M("01----") && M("-11---") && M("--11--") && M("---10-") && M("----01") && M("0-1---") && M("0--1--") && M("0---0-") && M("0----1") && M("-1-1--") && M("-1--0-") && M("-1---1") && M("--1-0-") && M("--1--1") && M("---1-1")) |
			// 011110:
			(M("01----") && M("-11---") && M("--11--") && M("---11-") && M("----10") && M("0-1---") && M("0--1--") && M("0---1-") && M("0----0") && M("-1-1--") && M("-1--1-") && M("-1---0") && M("--1-1-") && M("--1--0") && M("---1-0")) |
			// 011111:
			(M("01----") && M("-11---") && M("--11--") && M("---11-") && M("----11") && M("0-1---") && M("0--1--") && M("0---1-") && M("0----1") && M("-1-1--") && M("-1--1-") && M("-1---1") && M("--1-1-") && M("--1--1") && M("---1-1")) |


			// 100000:
			(M("10----") && M("-00---") && M("--00--") && M("---00-") && M("----00") && M("1-0---") && M("1--0--") && M("1---0-") && M("1----0") && M("-0-0--") && M("-0--0-") && M("-0---0") && M("--0-0-") && M("--0--0") && M("---0-0")) |
			// 100001:
			(M("10----") && M("-00---") && M("--00--") && M("---00-") && M("----01") && M("1-0---") && M("1--0--") && M("1---0-") && M("1----1") && M("-0-0--") && M("-0--0-") && M("-0---1") && M("--0-0-") && M("--0--1") && M("---0-1")) |
			// 100010:
			(M("10----") && M("-00---") && M("--00--") && M("---01-") && M("----10") && M("1-0---") && M("1--0--") && M("1---1-") && M("1----0") && M("-0-0--") && M("-0--1-") && M("-0---0") && M("--0-1-") && M("--0--0") && M("---0-0")) |
			// 100011:
			(M("10----") && M("-00---") && M("--00--") && M("---01-") && M("----11") && M("1-0---") && M("1--0--") && M("1---1-") && M("1----1") && M("-0-0--") && M("-0--1-") && M("-0---1") && M("--0-1-") && M("--0--1") && M("---0-1")) |
			// 100100:
			(M("10----") && M("-00---") && M("--01--") && M("---10-") && M("----00") && M("1-0---") && M("1--1--") && M("1---0-") && M("1----0") && M("-0-1--") && M("-0--0-") && M("-0---0") && M("--0-0-") && M("--0--0") && M("---1-0")) |
			// 100101:
			(M("10----") && M("-00---") && M("--01--") && M("---10-") && M("----01") && M("1-0---") && M("1--1--") && M("1---0-") && M("1----1") && M("-0-1--") && M("-0--0-") && M("-0---1") && M("--0-0-") && M("--0--1") && M("---1-1")) |
			// 100110:
			(M("10----") && M("-00---") && M("--01--") && M("---11-") && M("----10") && M("1-0---") && M("1--1--") && M("1---1-") && M("1----0") && M("-0-1--") && M("-0--1-") && M("-0---0") && M("--0-1-") && M("--0--0") && M("---1-0")) |
			// 100111:
			(M("10----") && M("-00---") && M("--01--") && M("---11-") && M("----11") && M("1-0---") && M("1--1--") && M("1---1-") && M("1----1") && M("-0-1--") && M("-0--1-") && M("-0---1") && M("--0-1-") && M("--0--1") && M("---1-1")) |

			// 101000:
			(M("10----") && M("-01---") && M("--10--") && M("---00-") && M("----00") && M("1-1---") && M("1--0--") && M("1---0-") && M("1----0") && M("-0-0--") && M("-0--0-") && M("-0---0") && M("--1-0-") && M("--1--0") && M("---0-0")) |
			// 101001:
			(M("10----") && M("-01---") && M("--10--") && M("---00-") && M("----01") && M("1-1---") && M("1--0--") && M("1---0-") && M("1----1") && M("-0-0--") && M("-0--0-") && M("-0---1") && M("--1-0-") && M("--1--1") && M("---0-1")) |
			// 101010:
			(M("10----") && M("-01---") && M("--10--") && M("---01-") && M("----10") && M("1-1---") && M("1--0--") && M("1---1-") && M("1----0") && M("-0-0--") && M("-0--1-") && M("-0---0") && M("--1-1-") && M("--1--0") && M("---0-0")) |
			// 101011:
			(M("10----") && M("-01---") && M("--10--") && M("---01-") && M("----11") && M("1-1---") && M("1--0--") && M("1---1-") && M("1----1") && M("-0-0--") && M("-0--1-") && M("-0---1") && M("--1-1-") && M("--1--1") && M("---0-1")) |
			// 101100:
			(M("10----") && M("-01---") && M("--11--") && M("---10-") && M("----00") && M("1-1---") && M("1--1--") && M("1---0-") && M("1----0") && M("-0-1--") && M("-0--0-") && M("-0---0") && M("--1-0-") && M("--1--0") && M("---1-0")) |
			// 101101:
			(M("10----") && M("-01---") && M("--11--") && M("---10-") && M("----01") && M("1-1---") && M("1--1--") && M("1---0-") && M("1----1") && M("-0-1--") && M("-0--0-") && M("-0---1") && M("--1-0-") && M("--1--1") && M("---1-1")) |
			// 101110:
			(M("10----") && M("-01---") && M("--11--") && M("---11-") && M("----10") && M("1-1---") && M("1--1--") && M("1---1-") && M("1----0") && M("-0-1--") && M("-0--1-") && M("-0---0") && M("--1-1-") && M("--1--0") && M("---1-0")) |
			// 101111:
			(M("10----") && M("-01---") && M("--11--") && M("---11-") && M("----11") && M("1-1---") && M("1--1--") && M("1---1-") && M("1----1") && M("-0-1--") && M("-0--1-") && M("-0---1") && M("--1-1-") && M("--1--1") && M("---1-1")) |

			// 110000:
			(M("11----") && M("-10---") && M("--00--") && M("---00-") && M("----00") && M("1-0---") && M("1--0--") && M("1---0-") && M("1----0") && M("-1-0--") && M("-1--0-") && M("-1---0") && M("--0-0-") && M("--0--0") && M("---0-0")) |
			// 110001:
			(M("11----") && M("-10---") && M("--00--") && M("---00-") && M("----01") && M("1-0---") && M("1--0--") && M("1---0-") && M("1----1") && M("-1-0--") && M("-1--0-") && M("-1---1") && M("--0-0-") && M("--0--1") && M("---0-1")) |
			// 110010:
			(M("11----") && M("-10---") && M("--00--") && M("---01-") && M("----10") && M("1-0---") && M("1--0--") && M("1---1-") && M("1----0") && M("-1-0--") && M("-1--1-") && M("-1---0") && M("--0-1-") && M("--0--0") && M("---0-0")) |
			// 110011:
			(M("11----") && M("-10---") && M("--00--") && M("---01-") && M("----11") && M("1-0---") && M("1--0--") && M("1---1-") && M("1----1") && M("-1-0--") && M("-1--1-") && M("-1---1") && M("--0-1-") && M("--0--1") && M("---0-1")) |
			// 110100:
			(M("11----") && M("-10---") && M("--01--") && M("---10-") && M("----00") && M("1-0---") && M("1--1--") && M("1---0-") && M("1----0") && M("-1-1--") && M("-1--0-") && M("-1---0") && M("--0-0-") && M("--0--0") && M("---1-0")) |
			// 110101:
			(M("11----") && M("-10---") && M("--01--") && M("---10-") && M("----01") && M("1-0---") && M("1--1--") && M("1---0-") && M("1----1") && M("-1-1--") && M("-1--0-") && M("-1---1") && M("--0-0-") && M("--0--1") && M("---1-1")) |
			// 110110:
			(M("11----") && M("-10---") && M("--01--") && M("---11-") && M("----10") && M("1-0---") && M("1--1--") && M("1---1-") && M("1----0") && M("-1-1--") && M("-1--1-") && M("-1---0") && M("--0-1-") && M("--0--0") && M("---1-0")) |
			// 110111:
			(M("11----") && M("-10---") && M("--01--") && M("---11-") && M("----11") && M("1-0---") && M("1--1--") && M("1---1-") && M("1----1") && M("-1-1--") && M("-1--1-") && M("-1---1") && M("--0-1-") && M("--0--1") && M("---1-1")) |

			// 111000:
			(M("11----") && M("-11---") && M("--10--") && M("---00-") && M("----00") && M("1-1---") && M("1--0--") && M("1---0-") && M("1----0") && M("-1-0--") && M("-1--0-") && M("-1---0") && M("--1-0-") && M("--1--0") && M("---0-0")) |
			// 111001:
			(M("11----") && M("-11---") && M("--10--") && M("---00-") && M("----01") && M("1-1---") && M("1--0--") && M("1---0-") && M("1----1") && M("-1-0--") && M("-1--0-") && M("-1---1") && M("--1-0-") && M("--1--1") && M("---0-1")) |
			// 111010:
			(M("11----") && M("-11---") && M("--10--") && M("---01-") && M("----10") && M("1-1---") && M("1--0--") && M("1---1-") && M("1----0") && M("-1-0--") && M("-1--1-") && M("-1---0") && M("--1-1-") && M("--1--0") && M("---0-0")) |
			// 111011:
			(M("11----") && M("-11---") && M("--10--") && M("---01-") && M("----11") && M("1-1---") && M("1--0--") && M("1---1-") && M("1----1") && M("-1-0--") && M("-1--1-") && M("-1---1") && M("--1-1-") && M("--1--1") && M("---0-1")) |
			// 111100:
			(M("11----") && M("-11---") && M("--11--") && M("---10-") && M("----00") && M("1-1---") && M("1--1--") && M("1---0-") && M("1----0") && M("-1-1--") && M("-1--0-") && M("-1---0") && M("--1-0-") && M("--1--0") && M("---1-0")) |
			// 111101:
			(M("11----") && M("-11---") && M("--11--") && M("---10-") && M("----01") && M("1-1---") && M("1--1--") && M("1---0-") && M("1----1") && M("-1-1--") && M("-1--0-") && M("-1---1") && M("--1-0-") && M("--1--1") && M("---1-1")) |
			// 111110:
			(M("11----") && M("-11---") && M("--11--") && M("---11-") && M("----10") && M("1-1---") && M("1--1--") && M("1---1-") && M("1----0") && M("-1-1--") && M("-1--1-") && M("-1---0") && M("--1-1-") && M("--1--0") && M("---1-0")) |
			// 111111:
			(M("11----") && M("-11---") && M("--11--") && M("---11-") && M("----11") && M("1-1---") && M("1--1--") && M("1---1-") && M("1----1") && M("-1-1--") && M("-1--1-") && M("-1---1") && M("--1-1-") && M("--1--1") && M("---1-1"));
	}

	printf("should not get here!\n");
	return false;
}

bool IsEqualOrIncluded(const char* Clause1, const char* Clause2)
{
	// NOTE: Clause1 is included in Clause 2 if: take Clause2 and arbitrarily
	// add '0'/'1' where Clause2 has '-' chars. C1 is not included in C2 if Clause1
	// has some '-' where Clause2 has '0' or '1'; C1 is also not included in C2 if
	// Clause1 is in conflict with Clause2.
	// 
	// Examples:
	// -included    : Clause1="00000-", Clause2="000---";
	// -not included: Clause1="-0000-", Clause2="000---".

	for (int m = 0; m < DigitNumber; m ++)
		if ((Clause1[m] != Clause2[m]) &&
			(Clause2[m] != '-'))
			return false;

	return true;
}

#define UNDERLYING_SOLUTION_NUMBER_MAX	65536

void DoConsistencyChecks(const int	DigitNumber, const int ClauseNumber,
						// const bool ActiveClauseTuple[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX])
						int			ActiveClauseTupleNumber,
						ACTStruct	ActiveClauseTuples[ACTIVE_CLAUSE_TUPLE_NUMBER_MAX])

{ // checks if the content of ActiveClauseTuple is as it should be, i.e. only tuples from Active CT Lines are left

	// IMPORTANT: this procedure needs exponentially much CPU time/memory,
	// so use it for verification purposes ONLY! The Polynomial Solver
	// does not necessarily need DoConsistencyChecks()!

	// assemble "resting ClauseTable Lines' Underlying Solutions"
	int			UnderlyingSolutionNumber = 0;
	static char	UnderlyingSolutions[UNDERLYING_SOLUTION_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];

	for (int m = 0; m < DigitNumber; m ++)
		UnderlyingSolutions[0][m] = '-'; // begin with pure '---...' and add clauses step by step
	UnderlyingSolutionNumber = 1;

	int			NewUnderlyingSolutionNumber = 0;
	static char	NewUnderlyingSolutions[UNDERLYING_SOLUTION_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];

	// NOTE: if the CNF is UNSAT, there will finally be *no*
	// Underlying Solution left, so UnderlyingSolutionNumber
	// will in this case be 0.

	for (int i = 0; i < PossibleClauseNumber; i ++)
	{
		if ((i % CombinationCount) == 0)
		{
			NewUnderlyingSolutionNumber = 0;
		}

		if (ExistsInCNF(i) == false)
		{
			for (int k = 0; k < UnderlyingSolutionNumber; k ++)
			{
				if (!IsInConflict(PossibleClauses[i], UnderlyingSolutions[k]))
				{
					// NOTE: we take UnderlyingSolutions[k] from last CT column and add current
					// clause PossibleClauses[i], as this clause and the Underlying Solution are
					// in some 'still true' CT line.

					for (int m = 0; m < DigitNumber; m ++)
					{
						if (PossibleClauses[i][m] != '-')
							NewUnderlyingSolutions[NewUnderlyingSolutionNumber][m] = PossibleClauses[i][m];
						else
							NewUnderlyingSolutions[NewUnderlyingSolutionNumber][m] = UnderlyingSolutions[k][m];
					}
					NewUnderlyingSolutions[NewUnderlyingSolutionNumber][DigitNumber] = '\0'; // add null termination to mark end of string

					if (NewUnderlyingSolutionNumber >= UNDERLYING_SOLUTION_NUMBER_MAX)
					{
						// If this happens, try increasing UNDERLYING_SOLUTION_NUMBER_MAX,
						// if not possible due to limited memory, implement a check that
						// avoids Underlying Solutions are added multiple times (this check
						// will probably need much CPU time), or use a smaller DigitNumber.

						printf("Error in DoConsistencyChecks(): too many UnderlyingSolutions!\n");
						getchar();
					}
					NewUnderlyingSolutionNumber ++;
				}
			}
		}

		if ((i % CombinationCount) == CombinationCount - 1)
		{
			// NOTE: if no clause has been added to some partial Underlying Solution,
			// because all suitable clauses that could have been added are thought
			// 'false' as they are existing in the SAT CNF, then throw away that
			// partial Underlying Solution.

			UnderlyingSolutionNumber = 0;
			for (int l = 0; l < NewUnderlyingSolutionNumber; l ++)
			{
				for (int m = 0; m <= DigitNumber; m ++)
					UnderlyingSolutions[UnderlyingSolutionNumber][m] = NewUnderlyingSolutions[l][m];
				UnderlyingSolutionNumber ++;
			}
		}
	}

	// check if TOO FEW ActiveClauseTuples found by Polynomial Solver
	for (int k = 0; k < UnderlyingSolutionNumber; k ++)
	{
		// NOTE: there must be a clause tuple (i,j) found for
		// i, j from each possible ClauseTable column tuple!
		// (That's why we use FoundAsi and FoundAsj.)

		/*
		bool FoundAsi = false;
		bool FoundAsj = false;

		for (int i = 0; i < PossibleClauseNumber; i ++)
		{
			if ((i % CombinationCount) == 0) // when i points to first clause of a ClauseTable column
				FoundAsi = false;

			if (!IsInConflict(
				PossibleClauses[i],
				UnderlyingSolutions[k]))
			{
				for (int j = 0; j < PossibleClauseNumber; j ++)
				{
					if (j / CombinationCount == i / CombinationCount)
						continue;

					if ((j % CombinationCount) == 0) // when j points to first clause of a ClauseTable column
						FoundAsj = false;

					if (!IsInConflict(
						PossibleClauses[j],
						UnderlyingSolutions[k]))
					{
						if (ActiveClauseTuple[i][j])
						{
							FoundAsi = true;
							FoundAsj = true;
						}
					}

					if ((j % CombinationCount) == CombinationCount - 1) // when j points to last clause of a ClauseTable column
					{
						if (FoundAsj == false)
						{
							printf("Error: ActiveClauseTuples should contain %s - missing tuple is %s,[some clause from block of %s]!\n", UnderlyingSolutions[k], PossibleClauses[i], PossibleClauses[j]);
							getchar();
						}
					}
				}
			}

			if ((i % CombinationCount) == CombinationCount - 1) // when i points to last clause of a ClauseTable column
			{
				if (FoundAsi == false)
				{
					printf("Error: ActiveClauseTuples should contain %s (no suitable i found)!\n", UnderlyingSolutions[k]);
					getchar();
				}
			}
		}
		*/

		bool FoundUS = false;

		for (int i = 0; i < ActiveClauseTupleNumber; i ++)
		{
			char Overlay[DIGIT_NUMBER_MAX + 1];
			ClauseOverlay(Overlay, ActiveClauseTuples[i].J, ActiveClauseTuples[i].K);
			if (IsEqualOrIncluded(
				UnderlyingSolutions[k],
				Overlay))
			{
				FoundUS = true;
				break;
			}
		}

		if (!FoundUS)
		{
			printf("Error: did not find Underlying Solution %s in ActiveClauseTuples!\n", UnderlyingSolutions[k]);
			getchar();
		}
	}

	// check if TOO MANY ActiveClauseTuples found by Polynomial Solver
	/*
	for (int i = 0; i < PossibleClauseNumber; i ++)
	{
		for (int j = 0; j < PossibleClauseNumber; j ++)
		{
			if (ActiveClauseTuple[i][j])
			{
			*/

	for (int i = 0; i < ActiveClauseTupleNumber; i ++)
	{
		char Overlay[DIGIT_NUMBER_MAX + 1];
		ClauseOverlay(Overlay, ActiveClauseTuples[i].J, ActiveClauseTuples[i].K);

				bool FoundTupleInUnderlyingSolution = false;

				for (int k = 0; k < UnderlyingSolutionNumber; k ++)
				{
					/*
					if (IsEqualOrIncluded(
						UnderlyingSolutions[k],
						PossibleClauses[i]) &&
						IsEqualOrIncluded(
						UnderlyingSolutions[k],
						PossibleClauses[j]))
					*/
					if (IsEqualOrIncluded(
						UnderlyingSolutions[k],
						Overlay))
					{
						FoundTupleInUnderlyingSolution = true;
						break;
					}
				}
				if (FoundTupleInUnderlyingSolution == false)
				{
					/*
					printf("Error: ActiveClauseTuple %s,%s is surplus!\n",
						PossibleClauses[i],
						PossibleClauses[j]);
					*/
					printf("Error: ActiveClauseTuple is surplus: %s\n", Overlay);
					getchar();
				}
			// }
		// }
	}
}

// ******* END OF NOT REQUIRED FOR ACTUAL SOLVER BUT FOR DEBUGGING ONLY *******
