// (c)2013-2015 by Louis Coder.

int					DigitNumberOld = 0;					// used to know when to (re-)initialize PossibleClauses[][]
int					DigitNumber;						// the clause size of any clause in the CNF, examples: 5 for e.g. "0-00-" or 7 for e.g. "1-0-0--"

int					SAT_TYPE_OLD = 0;					// used to detect changes
int					SAT_TYPE;							// 2 for 2-SAT or 3 for 3-SAT ('extern' as originally defined in Polynomial_3SAT_Solver.cpp)

#define				POSSIBLE_CLAUSE_NUMBER_MAX	(18500)	// highest PossibleClauseNumber supported by the solver (value gained through tests)

int					PossibleClauseNumber;				// depends from DigitNumber and the choice 2-SAT or 3-SAT
char				PossibleClauses[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];	// see PossibleClauses_Create()

#define				CT_DIGIT_NUMBER_MAX			(DIGIT_NUMBER_MAX + 1)
#define				CT_LINE_NUMBER_MAX			(4096)	// how many lines the ClauseTable can maximal have
#define				CT_COLUMN_NUMBER_MAX		(1024)	// how many columns the ClauseTable can maximal have

int					CTLineNumber = 0;					// the current count of lines in the ClauseTable
int					CTColumnNumber;						// the current count of columns in the ClauseTable
char				CT[CT_LINE_NUMBER_MAX][CT_COLUMN_NUMBER_MAX][CT_DIGIT_NUMBER_MAX + 1];	// contains Clauses; + 1 for null terminator

// **************************************************************

bool IsEqual(const char* Clause1, const char* Clause2)
{
	// NOTE: this function checks if the two char strings are identical.
	// DigitNumber is a global var initialized in CNF_Solve_Polynomial().

	for (int m = 0; m < DigitNumber; m ++)
		if ((Clause1[m] != Clause2[m]))
			return false;

	return true;
}

bool IsInConflict(const char* Clause1, const char* Clause2)
{
	// NOTE: this function checks if two clauses are in conflict.
	// Two clauses are in conflict if there's '0' at some position
	// in Clause1 where there's '1' in Clause2 (or vice versa).
	// 
	// For example, it is in conflict:
	// "0-0-" and "1-0-".
	// 
	// Not in conflict are for instance:
	// "0-0-" and "-00-", or
	// "0-0-" and "-0-0", or
	// "0-0-" and "-10-".

	for (int m = 0; m < DigitNumber; m ++)
		if ((Clause1[m] == '0' && Clause2[m] == '1') ||
			(Clause1[m] == '1' && Clause2[m] == '0'))
			return true;

	return false;
}

// **************************************************************

void ShowClause(char* Clause, bool NewLine = true)
{
	// For debugging.
	// You might use it like this:
	// ShowClause(PossibleClauses[x]);
	// or
	// ShowClause(&CNF[ClauseIndex * DigitNumber]);.

	if (Clause == NULL)
		return;

	for (int m = 0; m < DigitNumber; m ++)
		printf("%c", Clause[m]);
	if (NewLine)
		printf("\n");
}

// **************************************************************

void PossibleClauses_Create()
{
	PossibleClauseNumber = 0;

	if (SAT_TYPE == 2)
	{
		// if DigitNumber = 3:
		// 00-
		// 01-
		// 10-
		// 11-
		// 0-0
		// 0-1
		// 1-0
		// 1-1
		// -00
		// -01
		// -10
		// -11
		// PossibleClauseNumber = (3 * 4) = 12

		for (int D1 = 0; D1 < DigitNumber - 1; D1 ++) // position of first 0 or 1 digit in PossibleClause
		{
			for (int D2 = D1 + 1; D2 < DigitNumber; D2 ++) // position of second 0 or 1 digit in PossibleClause
			{
				for (int C = 0; C < 4; C ++) // CombinationIndex: 00, 01, 10, 11
				{
					// we arrive here O(DigitNumber^2) times!
					// So there are O(DigitNumber^2) PossibleClauses!

					for (int m = 0; m < DigitNumber; m ++)
						PossibleClauses[PossibleClauseNumber][m] = '-';

					PossibleClauses[PossibleClauseNumber][D1] = (C & 2 ? '1' : '0');
					PossibleClauses[PossibleClauseNumber][D2] = (C & 1 ? '1' : '0');
					PossibleClauses[PossibleClauseNumber][DigitNumber] = '\0';

					PossibleClauseNumber ++;

					if (PossibleClauseNumber > POSSIBLE_CLAUSE_NUMBER_MAX) // if this happens, try to increase POSSIBLE_CLAUSE_NUMBER_MAX and re-compile
					{
						printf("Warning: PossibleClauseNumber too large in PossibleClauses_Create() !");
						getchar();
					}
				}
			}
		}
	}

	if (SAT_TYPE == 3)
	{
		// if DigitNumber = 4:
		// 000-
		// 001-
		// 010-
		// 011-
		// 100-
		// 101-
		// 110-
		// 111-
		// 00-0
		// 00-1
		// 01-0
		// 01-1
		// 10-0
		// 10-1
		// 11-0
		// 11-1
		// 0-00
		// 0-01
		// 0-10
		// 0-11
		// 1-00
		// 1-01
		// 1-10
		// 1-11
		// -000
		// -001
		// -010
		// -011
		// -100
		// -101
		// -110
		// -111
		// PossibleClauseNumber = (4 * 8) = 32

		for (int D1 = 0; D1 < DigitNumber - 2; D1 ++) // position of first 0 or 1 digit in PossibleClause
		{
			for (int D2 = D1 + 1; D2 < DigitNumber - 1; D2 ++) // position of second 0 or 1 digit in PossibleClause
			{
				for (int D3 = D2 + 1; D3 < DigitNumber; D3 ++) // position of third 0 or 1 digit in PossibleClause
				{
					for (int C = 0; C < 8; C ++) // 000, 001, 010, 011, 100, 101, 110, 111
					{
						// we arrive here O(DigitNumber^3) times!
						// So there are O(DigitNumber^3) PossibleClauses!

						for (int m = 0; m < DigitNumber; m ++)
							PossibleClauses[PossibleClauseNumber][m] = '-';

						PossibleClauses[PossibleClauseNumber][D1] = (C & 4 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][D2] = (C & 2 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][D3] = (C & 1 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][DigitNumber] = '\0';

						PossibleClauseNumber ++;

						if (PossibleClauseNumber > POSSIBLE_CLAUSE_NUMBER_MAX) // if this happens, try to increase POSSIBLE_CLAUSE_NUMBER_MAX and re-compile
						{
							printf("Warning: PossibleClauseNumber too large in PossibleClauses_Create() !");
							getchar();
						}
					}
				}
			}
		}
	}
}

int GetPossibleClauseIndex(char* PossibleClause) // rather slow - use for debugging only!
{
	for (int i = 0; i < PossibleClauseNumber; i ++)
		if (PossibleClause[0] == PossibleClauses[i][0] &&
			PossibleClause[1] == PossibleClauses[i][1]) // quick incomplete check to increase speed
			if (IsEqual(PossibleClause, PossibleClauses[i]))
				return i;

	printf("ERROR in GetPossibleClauseIndex(): argument is not a PossibleClause or ClauseTable has not been created correctly !\n");
	getchar();
	return -1;
}

void ClauseTable_CreateDynamically() // fills CT[][] and sets CTLineNumber and CTColumnNumber
{
	CTLineNumber = 0;

	if (DigitNumber > CT_DIGIT_NUMBER_MAX)
	{
		printf("ERROR: ClauseTable_CreateDynamically() can only\n");
		printf("       be used for a DigitNumber <= %d !\n", CT_DIGIT_NUMBER_MAX);
		printf("       You tried to use it for DigitNumber = %d.\n", DigitNumber);
		getchar();
	}

	char CurrentLine[DIGIT_NUMBER_MAX + 1];

	for (int i = 0; i < DigitNumber; i ++)
		CurrentLine[i] = '0';

	// CurrentLine is initially (for DigitNumber = 5):
	// 00000.
	// In next while-iteration, it is:
	// 00001.
	// Then
	// 00010,
	// 00011,
	// 00100,
	// 00101,
	// 00110,
	// 00111, ...
	// Notice that this is EXPONENTIAL, but used for debugging/visualization ONLY!

	while (true)
	{
		for (int i = 0; i < CT_COLUMN_NUMBER_MAX; i ++)
			for (int m = 0; m < CT_DIGIT_NUMBER_MAX; m ++)
				CT[CTLineNumber][i][m] = '\0'; // clear

		int CTColumnIndex = 0;

		if (SAT_TYPE == 2)
		{
			for (int D1 = 0; D1 < DigitNumber - 1; D1 ++)
			{
				for (int D2 = D1 + 1; D2 < DigitNumber; D2 ++)
				{
					for (int m = 0; m < DigitNumber; m ++)
						CT[CTLineNumber][CTColumnIndex][m] = '-';
					
					CT[CTLineNumber][CTColumnIndex][D1] = CurrentLine[D1];
					CT[CTLineNumber][CTColumnIndex][D2] = CurrentLine[D2];
					CT[CTLineNumber][CTColumnIndex][DigitNumber] = '\0';

					CTColumnIndex ++;
				}
			}
		}

		if (SAT_TYPE == 3)
		{
			for (int D1 = 0; D1 < DigitNumber - 2; D1 ++)
			{
				for (int D2 = D1 + 1; D2 < DigitNumber - 1; D2 ++)
				{
					for (int D3 = D2 + 1; D3 < DigitNumber; D3 ++)
					{
						for (int m = 0; m < DigitNumber; m ++)
							CT[CTLineNumber][CTColumnIndex][m] = '-';

						CT[CTLineNumber][CTColumnIndex][D1] = CurrentLine[D1];
						CT[CTLineNumber][CTColumnIndex][D2] = CurrentLine[D2];
						CT[CTLineNumber][CTColumnIndex][D3] = CurrentLine[D3];
						CT[CTLineNumber][CTColumnIndex][DigitNumber] = '\0';

						CTColumnIndex ++;
					}
				}
			}
		}

		CTColumnNumber = CTColumnIndex;
		CTLineNumber ++;

		if (CTColumnNumber > CT_COLUMN_NUMBER_MAX)
		{
			printf("CTColumnNumber overflow !");
			getchar();
		}
		if (CTLineNumber > CT_LINE_NUMBER_MAX)
		{
			printf("CTLineNumber overflow !");
			getchar();
		}

		for (int m = DigitNumber - 1; m >= 0; m --)
			if (CurrentLine[m] == '0')
			{
				CurrentLine[m] = '1';
				goto NextLine;
			}
			else if (CurrentLine[m] == '1')
				CurrentLine[m] = '0';

		break;
NextLine:;
	}
}

void ClauseTable_Show()
{ // shows the whole ClauseTable (in the Windows console or, if specially compiled therefore, in the Linux shell)

#ifdef _WIN32
	SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_INTENSITY);
#endif
	printf("CLAUSE TABLE >>>\n\n");

	FILE* f = fopen("ClauseTable.txt", "wb");

	if (f == NULL)
		printf("warning: error opening file .\\ClauseTable.txt for writing!\n\n");

	fprintf(f, "\r\n");
	fprintf(f, "   ~~~ Project 'Polynomial 3-SAT Solver' by Matthias Mueller (www.louis-coder.com) ~~~   \r\n");
	fprintf(f, "\r\n");
	fprintf(f, "*****************************************************************************************\r\n");
	fprintf(f, "*                                                                                       *\r\n");
	fprintf(f, "* It follows the 'Clause Table' for %d-SAT, literal range %d.                           ", SAT_TYPE, DigitNumber); if (DigitNumber < 10) fprintf(f, "  *\r\n"); else fprintf(f, " *\r\n");
	fprintf(f, "*                                                                                       *\r\n");
	fprintf(f, "* When you replace each clause by 'true' if that clause does not exist in the SAT CNF   *\r\n");
	fprintf(f, "* to solve, or by 'false' if that clause is existing in the SAT CNF to solve, you get   *\r\n");
	fprintf(f, "* the solvability of the SAT CNF. If the Clause Table (DNF) as a whole is 'true', the   *\r\n");
	fprintf(f, "* SAT CNF has any solution (we do not know which one), if the Clause Table as a whole   *\r\n");
	fprintf(f, "* is 'false', the SAT CNF has no solution.                                              *\r\n");
	fprintf(f, "*                                                                                       *\r\n");
	fprintf(f, "* '&' means Boolean AND-ing and '|' means Boolean OR-ing (C/C++ notation).              *\r\n");
	fprintf(f, "*                                                                                       *\r\n");
	fprintf(f, "*****************************************************************************************\r\n");
	fprintf(f, "\r\n");
	fprintf(f, "\r\n");
	fprintf(f, "\r\n");

	for (int i = 0; i < CTLineNumber; i ++) // loop through the ClauseTable lines
	{
		printf("(");
		fprintf(f, "(");

		for (int j = 0; j < CTColumnNumber; j ++)
		{
			for (int m = 0; m < DigitNumber; m ++)
			{
				printf("%c", CT[i][j][m]);
				fprintf(f, "%c", CT[i][j][m]);
			}
#ifdef _WIN32
			SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE);
#endif
			if (j < CTColumnNumber - 1)
			{
				printf("  AND  ");
				fprintf(f, "  &  ");
			}
#ifdef _WIN32
			SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_INTENSITY);
#endif
		}
#ifdef _WIN32
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE);
#endif
		if (i < CTLineNumber - 1)
		{
			printf(") OR\n");
			fprintf(f, ") |\r\n");
		}
		else
		{
			printf(")\n");
			fprintf(f, ")\r\n");
		}
#ifdef _WIN32
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_INTENSITY);
#endif
	}

	fclose(f);

	printf("\n<<< END OF CLAUSE TABLE\n\n");
	printf("(Output also written to file 'ClauseTable.txt' in the current directory.)\n\n");

#ifdef _WIN32
	SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE);
#endif
}

// **************************************************************

void ClauseTable_Display_Core()
{
	ClauseTable_Show();
}

// **************************************************************

void ClauseTable_Display() // this is the one and only function to be used by the target project (i.e. here the procedure 'main()')
{
	// initialize some global arrays and vars
	if (DigitNumberOld != DigitNumber ||
		SAT_TYPE_OLD != SAT_TYPE) // at first run or when DigitNumber or/and SAT_TYPE changed
	{
		DigitNumberOld = DigitNumber;
		SAT_TYPE_OLD = SAT_TYPE;

		PossibleClauses_Create(); // initialize PossibleClauses[][] and PossibleClauseNumber
		ClauseTable_CreateDynamically(); // initialize the array CT[][] and set CTLineNumber and CTColumnNumber
	}

	// do the actual displaying
	ClauseTable_Display_Core();
}
